# Copyright 2020 Open Reaction Database Project Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Creates a Dataset by enumerating a template with a spreadsheet.

Usage:
    dataset_templating.py --template=<str> --spreadsheet=<str> --output=<str> [--no-validate]

Options:
    --template=<str>        Path to a Reaction pbtxt file defining a template
    --spreadsheet=<str>     Path to a spreadsheet file with a header row matching template placeholders
    --output=<str>          Filename for output Dataset
    --no-validate           If set, do not validate Reaction protos
"""
import docopt

from ord_schema.logging import get_logger
from ord_schema import message_helpers
from ord_schema import templating

logger = get_logger(__name__)


def main(kwargs):
    with open(kwargs["--template"]) as f:
        template_string = f.read()
    df = templating.read_spreadsheet(kwargs["--spreadsheet"])
    logger.info(
        "generating new Dataset from %s and %s",
        kwargs["--template"],
        kwargs["--spreadsheet"],
    )
    dataset = templating.generate_dataset(template_string, df, validate=not kwargs["--no-validate"])
    logger.info("writing new Dataset to %s", kwargs["--output"])
    message_helpers.write_message(dataset, kwargs["--output"])


if __name__ == "__main__":
    main(docopt.docopt(__doc__))



# Copyright 2020 The Open Reaction Database Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for ord_schema.frozen_storage."""

import dataclasses
import os
import tempfile

import pytest

from ord_schema import frozen_message
from ord_schema import message_helpers
from ord_schema.proto import reaction_pb2
from ord_schema.proto import test_pb2


def _freeze(message) -> frozen_message.FrozenMessage:
    """Runs a round-trip to disk as an extra check for FrozenMessage."""
    with tempfile.TemporaryDirectory() as tempdir:
        filename = os.path.join(tempdir, "message.pbtxt")
        message_helpers.write_message(message, filename)
        loaded_message = message_helpers.load_message(filename, type(message))
        return frozen_message.FrozenMessage(loaded_message)


def test_scalar_without_optional():
    """Illustrates the problem with non-optional scalar fields.

    This is what happens if scalar fields are not labeled `optional`:
      * Unset scalar fields return True with hasattr().
      * It's not possible to know whether a value matching the default was
        set explicitly.
    """
    frozen = _freeze(reaction_pb2.StirringConditions.StirringRate())
    assert hasattr(frozen, "rpm")  # Wrong!
    assert round(abs(frozen.rpm - 0.0), 7) == 0


def test_access_optional_scalar():
    frozen = _freeze(reaction_pb2.Percentage(value=12.3))
    assert hasattr(frozen, "value")
    assert round(abs(frozen.value - 12.3), 6) == 0
    assert not hasattr(frozen, "precision")
    with pytest.raises(AttributeError):
        _ = frozen.precision


def test_access_optional_bool():
    frozen = _freeze(test_pb2.Scalar(string_value="test"))
    assert not hasattr(frozen, "bool_value")
    with pytest.raises(AttributeError):
        _ = frozen.bool_value
    frozen = _freeze(test_pb2.Scalar(bool_value=False))
    assert hasattr(frozen, "bool_value")
    assert not frozen.bool_value
    frozen = _freeze(test_pb2.Scalar(bool_value=True))
    assert hasattr(frozen, "bool_value")
    assert frozen.bool_value


def test_access_submessage():
    frozen = _freeze(reaction_pb2.Reaction())
    assert not hasattr(frozen, "setup")
    with pytest.raises(AttributeError):
        _ = frozen.setup


def test_access_repeated_scalar():
    message = reaction_pb2.ProductMeasurement.MassSpecMeasurementDetails()
    frozen_empty = _freeze(message)
    assert len(frozen_empty.eic_masses) == 0
    with pytest.raises(IndexError):
        _ = frozen_empty.eic_masses[0]

    message.eic_masses.append(1.5)
    frozen = _freeze(message)
    assert len(frozen.eic_masses) == 1
    assert frozen.eic_masses[0] == 1.5
    with pytest.raises(IndexError):
        _ = frozen.eic_masses[1]


def test_access_repeated_submessage():
    message = reaction_pb2.Reaction()
    message.observations.add(comment="test")
    frozen = _freeze(message)
    assert len(frozen.observations) == 1
    assert frozen.observations[0].comment == "test"
    with pytest.raises(IndexError):
        _ = frozen.observations[1]
    assert len(frozen.outcomes) == 0
    with pytest.raises(IndexError):
        _ = frozen.outcomes[0]


def test_access_map_value():
    message = reaction_pb2.Reaction()
    message.inputs["test"].addition_order = 1
    frozen = _freeze(message)
    assert hasattr(frozen, "inputs")
    assert "test" in frozen.inputs
    assert hasattr(frozen.inputs["test"], "addition_order")
    assert frozen.inputs["test"].addition_order == 1
    assert "missing" not in frozen.inputs
    with pytest.raises(KeyError):
        _ = frozen.inputs["missing"]


def test_set_scalar_value():
    frozen = _freeze(reaction_pb2.StirringConditions.StirringRate())
    with pytest.raises(dataclasses.FrozenInstanceError):
        frozen.rpm = 123


def test_set_bool_value():
    frozen = _freeze(test_pb2.Scalar())
    with pytest.raises(dataclasses.FrozenInstanceError):
        frozen.bool_value = False


def test_set_submessage():
    frozen = _freeze(reaction_pb2.Reaction())
    with pytest.raises(AttributeError):
        frozen.setup.automation_platform = "test"
    with pytest.raises(AttributeError):
        frozen.setup.CopyFrom(reaction_pb2.ReactionSetup(automation_platform="test"))


def test_add_repeated_scalar():
    frozen = _freeze(reaction_pb2.ProductCompound())
    with pytest.raises(AttributeError):
        frozen.analysis_identity.append("test")


def test_add_repeated_submessage():
    frozen = _freeze(reaction_pb2.Reaction())
    with pytest.raises(AttributeError):
        frozen.observations.add(comment="test")


def test_set_map_value():
    frozen = _freeze(reaction_pb2.Reaction())
    with pytest.raises(KeyError):
        frozen.inputs["test"].addition_order = 1
    with pytest.raises(KeyError):
        frozen.inputs["test"].CopyFrom(reaction_pb2.ReactionInput(addition_order=1))


def test_modify_repeated_submessage():
    """See https://git.io/JfPf9."""
    message = reaction_pb2.Reaction()
    message.workups.add(type="ADDITION")
    frozen = _freeze(message)
    with pytest.raises(dataclasses.FrozenInstanceError):
        frozen.workups[0].type = reaction_pb2.ReactionWorkup.TEMPERATURE



# Copyright 2023 Open Reaction Database Project Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: ord-schema/proto/reaction.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1ford-schema/proto/reaction.proto\x12\x03ord\"\xd9\x03\n\x08Reaction\x12,\n\x0bidentifiers\x18\x01 \x03(\x0b\x32\x17.ord.ReactionIdentifier\x12)\n\x06inputs\x18\x02 \x03(\x0b\x32\x19.ord.Reaction.InputsEntry\x12!\n\x05setup\x18\x03 \x01(\x0b\x32\x12.ord.ReactionSetup\x12+\n\nconditions\x18\x04 \x01(\x0b\x32\x17.ord.ReactionConditions\x12!\n\x05notes\x18\x05 \x01(\x0b\x32\x12.ord.ReactionNotes\x12.\n\x0cobservations\x18\x06 \x03(\x0b\x32\x18.ord.ReactionObservation\x12$\n\x07workups\x18\x07 \x03(\x0b\x32\x13.ord.ReactionWorkup\x12&\n\x08outcomes\x18\x08 \x03(\x0b\x32\x14.ord.ReactionOutcome\x12+\n\nprovenance\x18\t \x01(\x0b\x32\x17.ord.ReactionProvenance\x12\x13\n\x0breaction_id\x18\n \x01(\t\x1a\x41\n\x0bInputsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12!\n\x05value\x18\x02 \x01(\x0b\x32\x12.ord.ReactionInput:\x02\x38\x01\"\xa7\x02\n\x12ReactionIdentifier\x12<\n\x04type\x18\x01 \x01(\x0e\x32..ord.ReactionIdentifier.ReactionIdentifierType\x12\x0f\n\x07\x64\x65tails\x18\x02 \x01(\t\x12\r\n\x05value\x18\x03 \x01(\t\x12\x16\n\tis_mapped\x18\x04 \x01(\x08H\x00\x88\x01\x01\"\x8c\x01\n\x16ReactionIdentifierType\x12\x0f\n\x0bUNSPECIFIED\x10\x00\x12\n\n\x06\x43USTOM\x10\x01\x12\x13\n\x0fREACTION_SMILES\x10\x02\x12\x15\n\x11REACTION_CXSMILES\x10\x06\x12\n\n\x06RDFILE\x10\x03\x12\n\n\x06RINCHI\x10\x04\x12\x11\n\rREACTION_TYPE\x10\x05\x42\x0c\n\n_is_mapped\"\xbc\x06\n\rReactionInput\x12!\n\ncomponents\x18\x01 \x03(\x0b\x32\r.ord.Compound\x12-\n\x10\x63rude_components\x18\x02 \x03(\x0b\x32\x13.ord.CrudeComponent\x12\x16\n\x0e\x61\x64\x64ition_order\x18\x03 \x01(\x05\x12 \n\raddition_time\x18\x04 \x01(\x0b\x32\t.ord.Time\x12\x38\n\x0e\x61\x64\x64ition_speed\x18\x05 \x01(\x0b\x32 .ord.ReactionInput.AdditionSpeed\x12$\n\x11\x61\x64\x64ition_duration\x18\x06 \x01(\x0b\x32\t.ord.Time\x12 \n\tflow_rate\x18\x07 \x01(\x0b\x32\r.ord.FlowRate\x12:\n\x0f\x61\x64\x64ition_device\x18\x08 \x01(\x0b\x32!.ord.ReactionInput.AdditionDevice\x12.\n\x14\x61\x64\x64ition_temperature\x18\t \x01(\x0b\x32\x10.ord.Temperature\x1a\xdc\x01\n\rAdditionSpeed\x12@\n\x04type\x18\x01 \x01(\x0e\x32\x32.ord.ReactionInput.AdditionSpeed.AdditionSpeedType\x12\x0f\n\x07\x64\x65tails\x18\x02 \x01(\t\"x\n\x11\x41\x64\x64itionSpeedType\x12\x0f\n\x0bUNSPECIFIED\x10\x00\x12\x0f\n\x0b\x41LL_AT_ONCE\x10\x01\x12\x08\n\x04\x46\x41ST\x10\x02\x12\x08\n\x04SLOW\x10\x03\x12\x0c\n\x08\x44ROPWISE\x10\x04\x12\x0e\n\nCONTINUOUS\x10\x05\x12\x0f\n\x0bPORTIONWISE\x10\x06\x1a\xd1\x01\n\x0e\x41\x64\x64itionDevice\x12\x42\n\x04type\x18\x01 \x01(\x0e\x32\x34.ord.ReactionInput.AdditionDevice.AdditionDeviceType\x12\x0f\n\x07\x64\x65tails\x18\x02 \x01(\t\"j\n\x12\x41\x64\x64itionDeviceType\x12\x0f\n\x0bUNSPECIFIED\x10\x00\x12\n\n\x06\x43USTOM\x10\x01\x12\x08\n\x04NONE\x10\x02\x12\x0b\n\x07SYRINGE\x10\x03\x12\x0b\n\x07\x43\x41NNULA\x10\x04\x12\x13\n\x0f\x41\x44\x44ITION_FUNNEL\x10\x05\"\xd6\x01\n\x06\x41mount\x12\x19\n\x04mass\x18\x01 \x01(\x0b\x32\t.ord.MassH\x00\x12\x1b\n\x05moles\x18\x02 \x01(\x0b\x32\n.ord.MolesH\x00\x12\x1d\n\x06volume\x18\x03 \x01(\x0b\x32\x0b.ord.VolumeH\x00\x12+\n\nunmeasured\x18\x05 \x01(\x0b\x32\x15.ord.UnmeasuredAmountH\x00\x12$\n\x17volume_includes_solutes\x18\x04 \x01(\x08H\x01\x88\x01\x01\x42\x06\n\x04kindB\x1a\n\x18_volume_includes_solutes\"\xbe\x01\n\x10UnmeasuredAmount\x12\x38\n\x04type\x18\x01 \x01(\x0e\x32*.ord.UnmeasuredAmount.UnmeasuredAmountType\x12\x0f\n\x07\x64\x65tails\x18\x02 \x01(\t\"_\n\x14UnmeasuredAmountType\x12\x0f\n\x0bUNSPECIFIED\x10\x00\x12\n\n\x06\x43USTOM\x10\x01\x12\r\n\tSATURATED\x10\x02\x12\r\n\tCATALYTIC\x10\x03\x12\x0c\n\x08TITRATED\x10\x04\"\xac\x01\n\x0e\x43rudeComponent\x12\x13\n\x0breaction_id\x18\x01 \x01(\t\x12\x1c\n\x0fincludes_workup\x18\x02 \x01(\x08H\x00\x88\x01\x01\x12\x1f\n\x12has_derived_amount\x18\x03 \x01(\x08H\x01\x88\x01\x01\x12\x1b\n\x06\x61mount\x18\x04 \x01(\x0b\x32\x0b.ord.AmountB\x12\n\x10_includes_workupB\x15\n\x13_has_derived_amount\"\xa5\x04\n\x08\x43ompound\x12,\n\x0bidentifiers\x18\x01 \x03(\x0b\x32\x17.ord.CompoundIdentifier\x12\x1b\n\x06\x61mount\x18\x02 \x01(\x0b\x32\x0b.ord.Amount\x12\x39\n\rreaction_role\x18\x03 \x01(\x0e\x32\".ord.ReactionRole.ReactionRoleType\x12\x18\n\x0bis_limiting\x18\x04 \x01(\x08H\x00\x88\x01\x01\x12.\n\x0cpreparations\x18\x05 \x03(\x0b\x32\x18.ord.CompoundPreparation\x12$\n\x06source\x18\x06 \x01(\x0b\x32\x14.ord.Compound.Source\x12-\n\x08\x66\x65\x61tures\x18\x07 \x03(\x0b\x32\x1b.ord.Compound.FeaturesEntry\x12-\n\x08\x61nalyses\x18\x08 \x03(\x0b\x32\x1b.ord.Compound.AnalysesEntry\x1a\x39\n\x06Source\x12\x0e\n\x06vendor\x18\x01 \x01(\t\x12\x12\n\ncatalog_id\x18\x02 \x01(\t\x12\x0b\n\x03lot\x18\x03 \x01(\t\x1a:\n\rFeaturesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x18\n\x05value\x18\x02 \x01(\x0b\x32\t.ord.Data:\x02\x38\x01\x1a>\n\rAnalysesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x1c\n\x05value\x18\x02 \x01(\x0b\x32\r.ord.Analysis:\x02\x38\x01\x42\x0e\n\x0c_is_limiting\"\xb2\x01\n\x0cReactionRole\"\xa1\x01\n\x10ReactionRoleType\x12\x0f\n\x0bUNSPECIFIED\x10\x00\x12\x0c\n\x08REACTANT\x10\x01\x12\x0b\n\x07REAGENT\x10\x02\x12\x0b\n\x07SOLVENT\x10\x03\x12\x0c\n\x08\x43\x41TALYST\x10\x04\x12\n\n\x06WORKUP\x10\x05\x12\x15\n\x11INTERNAL_STANDARD\x10\x06\x12\x16\n\x12\x41UTHENTIC_STANDARD\x10\x07\x12\x0b\n\x07PRODUCT\x10\x08\"\xf6\x01\n\x13\x43ompoundPreparation\x12>\n\x04type\x18\x01 \x01(\x0e\x32\x30.ord.CompoundPreparation.CompoundPreparationType\x12\x0f\n\x07\x64\x65tails\x18\x02 \x01(\t\x12\x13\n\x0breaction_id\x18\x03 \x01(\t\"y\n\x17\x43ompoundPreparationType\x12\x0f\n\x0bUNSPECIFIED\x10\x00\x12\n\n\x06\x43USTOM\x10\x01\x12\x08\n\x04NONE\x10\x02\x12\x0e\n\nREPURIFIED\x10\x03\x12\x0b\n\x07SPARGED\x10\x04\x12\t\n\x05\x44RIED\x10\x05\x12\x0f\n\x0bSYNTHESIZED\x10\x06\"\x82\x03\n\x12\x43ompoundIdentifier\x12<\n\x04type\x18\x01 \x01(\x0e\x32..ord.CompoundIdentifier.CompoundIdentifierType\x12\x0f\n\x07\x64\x65tails\x18\x02 \x01(\t\x12\r\n\x05value\x18\x03 \x01(\t\"\x8d\x02\n\x16\x43ompoundIdentifierType\x12\x0f\n\x0bUNSPECIFIED\x10\x00\x12\n\n\x06\x43USTOM\x10\x01\x12\n\n\x06SMILES\x10\x02\x12\t\n\x05INCHI\x10\x03\x12\x0c\n\x08MOLBLOCK\x10\x04\x12\x0e\n\nIUPAC_NAME\x10\x05\x12\x08\n\x04NAME\x10\x06\x12\x0e\n\nCAS_NUMBER\x10\x07\x12\x0f\n\x0bPUBCHEM_CID\x10\x08\x12\x11\n\rCHEMSPIDER_ID\x10\t\x12\x0c\n\x08\x43XSMILES\x10\n\x12\r\n\tINCHI_KEY\x10\x0b\x12\x07\n\x03XYZ\x10\x0c\x12\x0e\n\nUNIPROT_ID\x10\r\x12\n\n\x06PDB_ID\x10\x0e\x12\x17\n\x13\x41MINO_ACID_SEQUENCE\x10\x0f\x12\x08\n\x04HELM\x10\x10\"\xa7\x04\n\x06Vessel\x12$\n\x04type\x18\x01 \x01(\x0e\x32\x16.ord.Vessel.VesselType\x12\x0f\n\x07\x64\x65tails\x18\x02 \x01(\t\x12%\n\x08material\x18\x03 \x01(\x0b\x32\x13.ord.VesselMaterial\x12,\n\x0cpreparations\x18\x04 \x03(\x0b\x32\x16.ord.VesselPreparation\x12*\n\x0b\x61ttachments\x18\x05 \x03(\x0b\x32\x15.ord.VesselAttachment\x12\x1b\n\x06volume\x18\x06 \x01(\x0b\x32\x0b.ord.Volume\x12\x11\n\tvessel_id\x18\x07 \x01(\t\x12\x10\n\x08position\x18\x08 \x01(\t\x12\x0b\n\x03row\x18\t \x01(\t\x12\x0b\n\x03\x63ol\x18\n \x01(\t\"\x88\x02\n\nVesselType\x12\x0f\n\x0bUNSPECIFIED\x10\x00\x12\n\n\x06\x43USTOM\x10\x01\x12\x16\n\x12ROUND_BOTTOM_FLASK\x10\x02\x12\x08\n\x04VIAL\x10\x03\x12\x0e\n\nWELL_PLATE\x10\x04\x12\x12\n\x0eMICROWAVE_VIAL\x10\x05\x12\x08\n\x04TUBE\x10\x06\x12#\n\x1f\x43ONTINUOUS_STIRRED_TANK_REACTOR\x10\x07\x12\x16\n\x12PACKED_BED_REACTOR\x10\x08\x12\x0c\n\x08NMR_TUBE\x10\t\x12\x12\n\x0ePRESSURE_FLASK\x10\n\x12\x14\n\x10PRESSURE_REACTOR\x10\x0b\x12\x18\n\x14\x45LECTROCHEMICAL_CELL\x10\x0c\"\xcc\x01\n\x0eVesselMaterial\x12\x34\n\x04type\x18\x01 \x01(\x0e\x32&.ord.VesselMaterial.VesselMaterialType\x12\x0f\n\x07\x64\x65tails\x18\x02 \x01(\t\"s\n\x12VesselMaterialType\x12\x0f\n\x0bUNSPECIFIED\x10\x00\x12\n\n\x06\x43USTOM\x10\x01\x12\t\n\x05GLASS\x10\x02\x12\x11\n\rPOLYPROPYLENE\x10\x03\x12\x0b\n\x07PLASTIC\x10\x04\x12\t\n\x05METAL\x10\x05\x12\n\n\x06QUARTZ\x10\x06\"\x97\x03\n\x10VesselAttachment\x12\x38\n\x04type\x18\x01 \x01(\x0e\x32*.ord.VesselAttachment.VesselAttachmentType\x12\x0f\n\x07\x64\x65tails\x18\x02 \x01(\t\"\xb7\x02\n\x14VesselAttachmentType\x12\x0f\n\x0bUNSPECIFIED\x10\x00\x12\x08\n\x04NONE\x10\x01\x12\n\n\x06\x43USTOM\x10\x02\x12\n\n\x06SEPTUM\x10\x03\x12\x07\n\x03\x43\x41P\x10\x04\x12\x07\n\x03MAT\x10\x05\x12\x14\n\x10REFLUX_CONDENSER\x10\x06\x12\x0f\n\x0bVENT_NEEDLE\x10\x07\x12\x0e\n\nDEAN_STARK\x10\x08\x12\x0f\n\x0bVACUUM_TUBE\x10\t\x12\x13\n\x0f\x41\x44\x44ITION_FUNNEL\x10\n\x12\x0f\n\x0b\x44RYING_TUBE\x10\x0b\x12\x11\n\rALUMINUM_FOIL\x10\x0c\x12\x10\n\x0cTHERMOCOUPLE\x10\r\x12\x0b\n\x07\x42\x41LLOON\x10\x0e\x12\x0f\n\x0bGAS_ADAPTER\x10\x0f\x12\x16\n\x12PRESSURE_REGULATOR\x10\x10\x12\x11\n\rRELEASE_VALVE\x10\x11\"\xe8\x01\n\x11VesselPreparation\x12:\n\x04type\x18\x01 \x01(\x0e\x32,.ord.VesselPreparation.VesselPreparationType\x12\x0f\n\x07\x64\x65tails\x18\x02 \x01(\t\"\x85\x01\n\x15VesselPreparationType\x12\x0f\n\x0bUNSPECIFIED\x10\x00\x12\n\n\x06\x43USTOM\x10\x01\x12\x08\n\x04NONE\x10\x02\x12\x0e\n\nOVEN_DRIED\x10\x03\x12\x0f\n\x0b\x46LAME_DRIED\x10\x04\x12\x18\n\x14\x45VACUATED_BACKFILLED\x10\x05\x12\n\n\x06PURGED\x10\x06\"\xa0\x04\n\rReactionSetup\x12\x1b\n\x06vessel\x18\x01 \x01(\x0b\x32\x0b.ord.Vessel\x12\x19\n\x0cis_automated\x18\x02 \x01(\x08H\x00\x88\x01\x01\x12\x1b\n\x13\x61utomation_platform\x18\x03 \x01(\t\x12?\n\x0f\x61utomation_code\x18\x04 \x03(\x0b\x32&.ord.ReactionSetup.AutomationCodeEntry\x12;\n\x0b\x65nvironment\x18\x05 \x01(\x0b\x32&.ord.ReactionSetup.ReactionEnvironment\x1a@\n\x13\x41utomationCodeEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x18\n\x05value\x18\x02 \x01(\x0b\x32\t.ord.Data:\x02\x38\x01\x1a\xe8\x01\n\x13ReactionEnvironment\x12L\n\x04type\x18\x01 \x01(\x0e\x32>.ord.ReactionSetup.ReactionEnvironment.ReactionEnvironmentType\x12\x0f\n\x07\x64\x65tails\x18\x02 \x01(\t\"r\n\x17ReactionEnvironmentType\x12\x0f\n\x0bUNSPECIFIED\x10\x00\x12\n\n\x06\x43USTOM\x10\x01\x12\r\n\tFUME_HOOD\x10\x02\x12\r\n\tBENCH_TOP\x10\x03\x12\r\n\tGLOVE_BOX\x10\x04\x12\r\n\tGLOVE_BAG\x10\x05\x42\x0f\n\r_is_automated\"\xb5\x03\n\x12ReactionConditions\x12/\n\x0btemperature\x18\x01 \x01(\x0b\x32\x1a.ord.TemperatureConditions\x12)\n\x08pressure\x18\x02 \x01(\x0b\x32\x17.ord.PressureConditions\x12)\n\x08stirring\x18\x03 \x01(\x0b\x32\x17.ord.StirringConditions\x12\x31\n\x0cillumination\x18\x04 \x01(\x0b\x32\x1b.ord.IlluminationConditions\x12\x39\n\x10\x65lectrochemistry\x18\x05 \x01(\x0b\x32\x1f.ord.ElectrochemistryConditions\x12!\n\x04\x66low\x18\x06 \x01(\x0b\x32\x13.ord.FlowConditions\x12\x13\n\x06reflux\x18\x07 \x01(\x08H\x00\x88\x01\x01\x12\x0f\n\x02ph\x18\x08 \x01(\x02H\x01\x88\x01\x01\x12#\n\x16\x63onditions_are_dynamic\x18\t \x01(\x08H\x02\x88\x01\x01\x12\x0f\n\x07\x64\x65tails\x18\n \x01(\tB\t\n\x07_refluxB\x05\n\x03_phB\x19\n\x17_conditions_are_dynamic\"\xe2\x06\n\x15TemperatureConditions\x12>\n\x07\x63ontrol\x18\x01 \x01(\x0b\x32-.ord.TemperatureConditions.TemperatureControl\x12\"\n\x08setpoint\x18\x02 \x01(\x0b\x32\x10.ord.Temperature\x12G\n\x0cmeasurements\x18\x03 \x03(\x0b\x32\x31.ord.TemperatureConditions.TemperatureMeasurement\x1a\xd4\x02\n\x12TemperatureControl\x12R\n\x04type\x18\x01 \x01(\x0e\x32\x44.ord.TemperatureConditions.TemperatureControl.TemperatureControlType\x12\x0f\n\x07\x64\x65tails\x18\x02 \x01(\t\"\xd8\x01\n\x16TemperatureControlType\x12\x0f\n\x0bUNSPECIFIED\x10\x00\x12\n\n\x06\x43USTOM\x10\x01\x12\x0b\n\x07\x41MBIENT\x10\x02\x12\x0c\n\x08OIL_BATH\x10\x03\x12\x0e\n\nWATER_BATH\x10\x04\x12\r\n\tSAND_BATH\x10\x05\x12\x0c\n\x08ICE_BATH\x10\x06\x12\x16\n\x12\x44RY_ALUMINUM_PLATE\x10\x07\x12\r\n\tMICROWAVE\x10\x08\x12\x10\n\x0c\x44RY_ICE_BATH\x10\t\x12\x0b\n\x07\x41IR_FAN\x10\n\x12\x13\n\x0fLIQUID_NITROGEN\x10\x0b\x1a\xc4\x02\n\x16TemperatureMeasurement\x12Z\n\x04type\x18\x01 \x01(\x0e\x32L.ord.TemperatureConditions.TemperatureMeasurement.TemperatureMeasurementType\x12\x0f\n\x07\x64\x65tails\x18\x02 \x01(\t\x12\x17\n\x04time\x18\x03 \x01(\x0b\x32\t.ord.Time\x12%\n\x0btemperature\x18\x04 \x01(\x0b\x32\x10.ord.Temperature\"}\n\x1aTemperatureMeasurementType\x12\x0f\n\x0bUNSPECIFIED\x10\x00\x12\n\n\x06\x43USTOM\x10\x01\x12\x19\n\x15THERMOCOUPLE_INTERNAL\x10\x02\x12\x19\n\x15THERMOCOUPLE_EXTERNAL\x10\x03\x12\x0c\n\x08INFRARED\x10\x04\"\x8c\x08\n\x12PressureConditions\x12\x38\n\x07\x63ontrol\x18\x01 \x01(\x0b\x32\'.ord.PressureConditions.PressureControl\x12\x1f\n\x08setpoint\x18\x02 \x01(\x0b\x32\r.ord.Pressure\x12\x36\n\natmosphere\x18\x03 \x01(\x0b\x32\".ord.PressureConditions.Atmosphere\x12\x41\n\x0cmeasurements\x18\x04 \x03(\x0b\x32+.ord.PressureConditions.PressureMeasurement\x1a\xe0\x01\n\x0fPressureControl\x12I\n\x04type\x18\x01 \x01(\x0e\x32;.ord.PressureConditions.PressureControl.PressureControlType\x12\x0f\n\x07\x64\x65tails\x18\x02 \x01(\t\"q\n\x13PressureControlType\x12\x0f\n\x0bUNSPECIFIED\x10\x00\x12\n\n\x06\x43USTOM\x10\x01\x12\x0b\n\x07\x41MBIENT\x10\x02\x12\x13\n\x0fSLIGHT_POSITIVE\x10\x03\x12\n\n\x06SEALED\x10\x04\x12\x0f\n\x0bPRESSURIZED\x10\x05\x1a\xb5\x02\n\nAtmosphere\x12?\n\x04type\x18\x01 \x01(\x0e\x32\x31.ord.PressureConditions.Atmosphere.AtmosphereType\x12\x0f\n\x07\x64\x65tails\x18\x02 \x01(\t\"\xd4\x01\n\x0e\x41tmosphereType\x12\x0f\n\x0bUNSPECIFIED\x10\x00\x12\n\n\x06\x43USTOM\x10\x01\x12\x07\n\x03\x41IR\x10\x02\x12\x0c\n\x08NITROGEN\x10\x03\x12\t\n\x05\x41RGON\x10\x04\x12\n\n\x06OXYGEN\x10\x05\x12\x0c\n\x08HYDROGEN\x10\x06\x12\x13\n\x0f\x43\x41RBON_MONOXIDE\x10\x07\x12\x12\n\x0e\x43\x41RBON_DIOXIDE\x10\x08\x12\x0b\n\x07METHANE\x10\t\x12\x0b\n\x07\x41MMONIA\x10\n\x12\t\n\x05OZONE\x10\x0b\x12\x0c\n\x08\x45THYLENE\x10\x0c\x12\r\n\tACETYLENE\x10\r\x1a\x84\x02\n\x13PressureMeasurement\x12Q\n\x04type\x18\x01 \x01(\x0e\x32\x43.ord.PressureConditions.PressureMeasurement.PressureMeasurementType\x12\x0f\n\x07\x64\x65tails\x18\x02 \x01(\t\x12\x17\n\x04time\x18\x03 \x01(\x0b\x32\t.ord.Time\x12\x1f\n\x08pressure\x18\x04 \x01(\x0b\x32\r.ord.Pressure\"O\n\x17PressureMeasurementType\x12\x0f\n\x0bUNSPECIFIED\x10\x00\x12\n\n\x06\x43USTOM\x10\x01\x12\x17\n\x13PRESSURE_TRANSDUCER\x10\x02\"\xdc\x03\n\x12StirringConditions\x12\x38\n\x04type\x18\x01 \x01(\x0e\x32*.ord.StirringConditions.StirringMethodType\x12\x0f\n\x07\x64\x65tails\x18\x02 \x01(\t\x12\x32\n\x04rate\x18\x03 \x01(\x0b\x32$.ord.StirringConditions.StirringRate\x1a\xb5\x01\n\x0cStirringRate\x12\x43\n\x04type\x18\x01 \x01(\x0e\x32\x35.ord.StirringConditions.StirringRate.StirringRateType\x12\x0f\n\x07\x64\x65tails\x18\x02 \x01(\t\x12\x0b\n\x03rpm\x18\x03 \x01(\x05\"B\n\x10StirringRateType\x12\x0f\n\x0bUNSPECIFIED\x10\x00\x12\x08\n\x04HIGH\x10\x01\x12\n\n\x06MEDIUM\x10\x02\x12\x07\n\x03LOW\x10\x03\"\x8e\x01\n\x12StirringMethodType\x12\x0f\n\x0bUNSPECIFIED\x10\x00\x12\n\n\x06\x43USTOM\x10\x01\x12\x08\n\x04NONE\x10\x02\x12\x0c\n\x08STIR_BAR\x10\x03\x12\x12\n\x0eOVERHEAD_MIXER\x10\x04\x12\r\n\tAGITATION\x10\x05\x12\x10\n\x0c\x42\x41LL_MILLING\x10\x06\x12\x0e\n\nSONICATION\x10\x07\"\xe8\x02\n\x16IlluminationConditions\x12:\n\x04type\x18\x01 \x01(\x0e\x32,.ord.IlluminationConditions.IlluminationType\x12\x0f\n\x07\x64\x65tails\x18\x02 \x01(\t\x12(\n\x0fpeak_wavelength\x18\x03 \x01(\x0b\x32\x0f.ord.Wavelength\x12\r\n\x05\x63olor\x18\x04 \x01(\t\x12\'\n\x12\x64istance_to_vessel\x18\x05 \x01(\x0b\x32\x0b.ord.Length\"\x9e\x01\n\x10IlluminationType\x12\x0f\n\x0bUNSPECIFIED\x10\x00\x12\n\n\x06\x43USTOM\x10\x01\x12\x0b\n\x07\x41MBIENT\x10\x02\x12\x08\n\x04\x44\x41RK\x10\x03\x12\x07\n\x03LED\x10\x04\x12\x10\n\x0cHALOGEN_LAMP\x10\x05\x12\x12\n\x0e\x44\x45UTERIUM_LAMP\x10\x06\x12\x13\n\x0fSOLAR_SIMULATOR\x10\x07\x12\x12\n\x0e\x42ROAD_SPECTRUM\x10\x08\"\xe0\x06\n\x1a\x45lectrochemistryConditions\x12\x42\n\x04type\x18\x01 \x01(\x0e\x32\x34.ord.ElectrochemistryConditions.ElectrochemistryType\x12\x0f\n\x07\x64\x65tails\x18\x02 \x01(\t\x12\x1d\n\x07\x63urrent\x18\x03 \x01(\x0b\x32\x0c.ord.Current\x12\x1d\n\x07voltage\x18\x04 \x01(\x0b\x32\x0c.ord.Voltage\x12\x16\n\x0e\x61node_material\x18\x05 \x01(\t\x12\x18\n\x10\x63\x61thode_material\x18\x06 \x01(\t\x12)\n\x14\x65lectrode_separation\x18\x07 \x01(\x0b\x32\x0b.ord.Length\x12Q\n\x0cmeasurements\x18\x08 \x03(\x0b\x32;.ord.ElectrochemistryConditions.ElectrochemistryMeasurement\x12\x42\n\x04\x63\x65ll\x18\t \x01(\x0b\x32\x34.ord.ElectrochemistryConditions.ElectrochemistryCell\x1at\n\x1b\x45lectrochemistryMeasurement\x12\x17\n\x04time\x18\x01 \x01(\x0b\x32\t.ord.Time\x12\x1d\n\x07\x63urrent\x18\x02 \x01(\x0b\x32\x0c.ord.Current\x12\x1d\n\x07voltage\x18\x03 \x01(\x0b\x32\x0c.ord.Voltage\x1a\xe3\x01\n\x14\x45lectrochemistryCell\x12[\n\x04type\x18\x01 \x01(\x0e\x32M.ord.ElectrochemistryConditions.ElectrochemistryCell.ElectrochemistryCellType\x12\x0f\n\x07\x64\x65tails\x18\x02 \x01(\t\"]\n\x18\x45lectrochemistryCellType\x12\x0f\n\x0bUNSPECIFIED\x10\x00\x12\n\n\x06\x43USTOM\x10\x01\x12\x10\n\x0c\x44IVIDED_CELL\x10\x02\x12\x12\n\x0eUNDIVIDED_CELL\x10\x03\"_\n\x14\x45lectrochemistryType\x12\x0f\n\x0bUNSPECIFIED\x10\x00\x12\n\n\x06\x43USTOM\x10\x01\x12\x14\n\x10\x43ONSTANT_CURRENT\x10\x02\x12\x14\n\x10\x43ONSTANT_VOLTAGE\x10\x03\"\x94\x04\n\x0e\x46lowConditions\x12*\n\x04type\x18\x01 \x01(\x0e\x32\x1c.ord.FlowConditions.FlowType\x12\x0f\n\x07\x64\x65tails\x18\x02 \x01(\t\x12\x11\n\tpump_type\x18\x03 \x01(\t\x12*\n\x06tubing\x18\x04 \x01(\x0b\x32\x1a.ord.FlowConditions.Tubing\x1a\x88\x02\n\x06Tubing\x12\x33\n\x04type\x18\x01 \x01(\x0e\x32%.ord.FlowConditions.Tubing.TubingType\x12\x0f\n\x07\x64\x65tails\x18\x02 \x01(\t\x12\x1d\n\x08\x64iameter\x18\x03 \x01(\x0b\x32\x0b.ord.Length\"\x98\x01\n\nTubingType\x12\x0f\n\x0bUNSPECIFIED\x10\x00\x12\n\n\x06\x43USTOM\x10\x01\x12\t\n\x05STEEL\x10\x02\x12\n\n\x06\x43OPPER\x10\x03\x12\x07\n\x03PFA\x10\x04\x12\x07\n\x03\x46\x45P\x10\x05\x12\x0c\n\x08TEFLONAF\x10\x06\x12\x08\n\x04PTFE\x10\x07\x12\t\n\x05GLASS\x10\x08\x12\n\n\x06QUARTZ\x10\t\x12\x0b\n\x07SILICON\x10\n\x12\x08\n\x04PDMS\x10\x0b\"{\n\x08\x46lowType\x12\x0f\n\x0bUNSPECIFIED\x10\x00\x12\n\n\x06\x43USTOM\x10\x01\x12\x15\n\x11PLUG_FLOW_REACTOR\x10\x02\x12#\n\x1f\x43ONTINUOUS_STIRRED_TANK_REACTOR\x10\x03\x12\x16\n\x12PACKED_BED_REACTOR\x10\x04\"\xc0\x03\n\rReactionNotes\x12\x1d\n\x10is_heterogeneous\x18\x01 \x01(\x08H\x00\x88\x01\x01\x12\x1e\n\x11\x66orms_precipitate\x18\x02 \x01(\x08H\x01\x88\x01\x01\x12\x1a\n\ris_exothermic\x18\x03 \x01(\x08H\x02\x88\x01\x01\x12\x16\n\toffgasses\x18\x04 \x01(\x08H\x03\x88\x01\x01\x12%\n\x18is_sensitive_to_moisture\x18\x05 \x01(\x08H\x04\x88\x01\x01\x12#\n\x16is_sensitive_to_oxygen\x18\x06 \x01(\x08H\x05\x88\x01\x01\x12\"\n\x15is_sensitive_to_light\x18\x07 \x01(\x08H\x06\x88\x01\x01\x12\x14\n\x0csafety_notes\x18\x08 \x01(\t\x12\x19\n\x11procedure_details\x18\t \x01(\tB\x13\n\x11_is_heterogeneousB\x14\n\x12_forms_precipitateB\x10\n\x0e_is_exothermicB\x0c\n\n_offgassesB\x1b\n\x19_is_sensitive_to_moistureB\x19\n\x17_is_sensitive_to_oxygenB\x18\n\x16_is_sensitive_to_light\"Y\n\x13ReactionObservation\x12\x17\n\x04time\x18\x01 \x01(\x0b\x32\t.ord.Time\x12\x0f\n\x07\x63omment\x18\x02 \x01(\t\x12\x18\n\x05image\x18\x03 \x01(\x0b\x32\t.ord.Data\"\xcb\x05\n\x0eReactionWorkup\x12\x34\n\x04type\x18\x01 \x01(\x0e\x32&.ord.ReactionWorkup.ReactionWorkupType\x12\x0f\n\x07\x64\x65tails\x18\x02 \x01(\t\x12\x1b\n\x08\x64uration\x18\x03 \x01(\x0b\x32\t.ord.Time\x12!\n\x05input\x18\x04 \x01(\x0b\x32\x12.ord.ReactionInput\x12\x1b\n\x06\x61mount\x18\x05 \x01(\x0b\x32\x0b.ord.Amount\x12/\n\x0btemperature\x18\x06 \x01(\x0b\x32\x1a.ord.TemperatureConditions\x12\x12\n\nkeep_phase\x18\x07 \x01(\t\x12)\n\x08stirring\x18\x08 \x01(\x0b\x32\x17.ord.StirringConditions\x12\x16\n\ttarget_ph\x18\t \x01(\x02H\x00\x88\x01\x01\x12\x19\n\x0cis_automated\x18\n \x01(\x08H\x01\x88\x01\x01\"\xd2\x02\n\x12ReactionWorkupType\x12\x0f\n\x0bUNSPECIFIED\x10\x00\x12\n\n\x06\x43USTOM\x10\x01\x12\x0c\n\x08\x41\x44\x44ITION\x10\x02\x12\x0b\n\x07\x41LIQUOT\x10\x03\x12\x0f\n\x0bTEMPERATURE\x10\x04\x12\x11\n\rCONCENTRATION\x10\x05\x12\x0e\n\nEXTRACTION\x10\x06\x12\x0e\n\nFILTRATION\x10\x07\x12\x08\n\x04WASH\x10\x08\x12\x11\n\rDRY_IN_VACUUM\x10\t\x12\x15\n\x11\x44RY_WITH_MATERIAL\x10\n\x12\x18\n\x14\x46LASH_CHROMATOGRAPHY\x10\x0b\x12\x18\n\x14OTHER_CHROMATOGRAPHY\x10\x0c\x12\x0e\n\nSCAVENGING\x10\r\x12\x08\n\x04WAIT\x10\x0e\x12\x0c\n\x08STIRRING\x10\x0f\x12\r\n\tPH_ADJUST\x10\x10\x12\x0f\n\x0b\x44ISSOLUTION\x10\x11\x12\x10\n\x0c\x44ISTILLATION\x10\x12\x42\x0c\n\n_target_phB\x0f\n\r_is_automated\"\xf6\x01\n\x0fReactionOutcome\x12 \n\rreaction_time\x18\x01 \x01(\x0b\x32\t.ord.Time\x12#\n\nconversion\x18\x02 \x01(\x0b\x32\x0f.ord.Percentage\x12&\n\x08products\x18\x03 \x03(\x0b\x32\x14.ord.ProductCompound\x12\x34\n\x08\x61nalyses\x18\x04 \x03(\x0b\x32\".ord.ReactionOutcome.AnalysesEntry\x1a>\n\rAnalysesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x1c\n\x05value\x18\x02 \x01(\x0b\x32\r.ord.Analysis:\x02\x38\x01\"\x8d\x05\n\x0fProductCompound\x12,\n\x0bidentifiers\x18\x01 \x03(\x0b\x32\x17.ord.CompoundIdentifier\x12\x1f\n\x12is_desired_product\x18\x02 \x01(\x08H\x00\x88\x01\x01\x12-\n\x0cmeasurements\x18\x03 \x03(\x0b\x32\x17.ord.ProductMeasurement\x12\x16\n\x0eisolated_color\x18\x04 \x01(\t\x12-\n\x07texture\x18\x05 \x01(\x0b\x32\x1c.ord.ProductCompound.Texture\x12\x34\n\x08\x66\x65\x61tures\x18\x06 \x03(\x0b\x32\".ord.ProductCompound.FeaturesEntry\x12\x39\n\rreaction_role\x18\x07 \x01(\x0e\x32\".ord.ReactionRole.ReactionRoleType\x1a\xf0\x01\n\x07Texture\x12\x36\n\x04type\x18\x01 \x01(\x0e\x32(.ord.ProductCompound.Texture.TextureType\x12\x0f\n\x07\x64\x65tails\x18\x02 \x01(\t\"\x9b\x01\n\x0bTextureType\x12\x0f\n\x0bUNSPECIFIED\x10\x00\x12\n\n\x06\x43USTOM\x10\x01\x12\n\n\x06POWDER\x10\x02\x12\x0b\n\x07\x43RYSTAL\x10\x03\x12\x07\n\x03OIL\x10\x04\x12\x13\n\x0f\x41MORPHOUS_SOLID\x10\x05\x12\x08\n\x04\x46OAM\x10\x06\x12\x07\n\x03WAX\x10\x07\x12\x0e\n\nSEMI_SOLID\x10\x08\x12\t\n\x05SOLID\x10\t\x12\n\n\x06LIQUID\x10\n\x1a:\n\rFeaturesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x18\n\x05value\x18\x02 \x01(\x0b\x32\t.ord.Data:\x02\x38\x01\x42\x15\n\x13_is_desired_product\"\xfb\n\n\x12ProductMeasurement\x12\x14\n\x0c\x61nalysis_key\x18\x01 \x01(\t\x12<\n\x04type\x18\x02 \x01(\x0e\x32..ord.ProductMeasurement.ProductMeasurementType\x12\x0f\n\x07\x64\x65tails\x18\x03 \x01(\t\x12#\n\x16uses_internal_standard\x18\x04 \x01(\x08H\x01\x88\x01\x01\x12\x1a\n\ris_normalized\x18\x05 \x01(\x08H\x02\x88\x01\x01\x12$\n\x17uses_authentic_standard\x18\x06 \x01(\x08H\x03\x88\x01\x01\x12)\n\x12\x61uthentic_standard\x18\x07 \x01(\x0b\x32\r.ord.Compound\x12%\n\npercentage\x18\x08 \x01(\x0b\x32\x0f.ord.PercentageH\x00\x12&\n\x0b\x66loat_value\x18\t \x01(\x0b\x32\x0f.ord.FloatValueH\x00\x12\x16\n\x0cstring_value\x18\n \x01(\tH\x00\x12\x1d\n\x06\x61mount\x18\x0b \x01(\x0b\x32\x0b.ord.AmountH\x00\x12!\n\x0eretention_time\x18\x0c \x01(\x0b\x32\t.ord.Time\x12M\n\x11mass_spec_details\x18\r \x01(\x0b\x32\x32.ord.ProductMeasurement.MassSpecMeasurementDetails\x12\x38\n\x0bselectivity\x18\x0e \x01(\x0b\x32#.ord.ProductMeasurement.Selectivity\x12#\n\nwavelength\x18\x0f \x01(\x0b\x32\x0f.ord.Wavelength\x1a\xe9\x02\n\x1aMassSpecMeasurementDetails\x12X\n\x04type\x18\x01 \x01(\x0e\x32J.ord.ProductMeasurement.MassSpecMeasurementDetails.MassSpecMeasurementType\x12\x0f\n\x07\x64\x65tails\x18\x02 \x01(\t\x12\x1b\n\x0etic_minimum_mz\x18\x03 \x01(\x02H\x00\x88\x01\x01\x12\x1b\n\x0etic_maximum_mz\x18\x04 \x01(\x02H\x01\x88\x01\x01\x12\x12\n\neic_masses\x18\x05 \x03(\x02\"l\n\x17MassSpecMeasurementType\x12\x0f\n\x0bUNSPECIFIED\x10\x00\x12\n\n\x06\x43USTOM\x10\x01\x12\x07\n\x03TIC\x10\x02\x12\x10\n\x0cTIC_POSITIVE\x10\x03\x12\x10\n\x0cTIC_NEGATIVE\x10\x04\x12\x07\n\x03\x45IC\x10\x05\x42\x11\n\x0f_tic_minimum_mzB\x11\n\x0f_tic_maximum_mz\x1a\xb9\x01\n\x0bSelectivity\x12\x41\n\x04type\x18\x01 \x01(\x0e\x32\x33.ord.ProductMeasurement.Selectivity.SelectivityType\x12\x0f\n\x07\x64\x65tails\x18\x02 \x01(\t\"V\n\x0fSelectivityType\x12\x0f\n\x0bUNSPECIFIED\x10\x00\x12\n\n\x06\x43USTOM\x10\x01\x12\x06\n\x02\x45\x45\x10\x02\x12\x06\n\x02\x45R\x10\x03\x12\x06\n\x02\x44R\x10\x04\x12\x06\n\x02\x45Z\x10\x05\x12\x06\n\x02ZE\x10\x06\"\x9c\x01\n\x16ProductMeasurementType\x12\x0f\n\x0bUNSPECIFIED\x10\x00\x12\n\n\x06\x43USTOM\x10\x01\x12\x0c\n\x08IDENTITY\x10\x02\x12\t\n\x05YIELD\x10\x03\x12\x0f\n\x0bSELECTIVITY\x10\x04\x12\n\n\x06PURITY\x10\x05\x12\x08\n\x04\x41REA\x10\x06\x12\n\n\x06\x43OUNTS\x10\x07\x12\r\n\tINTENSITY\x10\x08\x12\n\n\x06\x41MOUNT\x10\tB\x07\n\x05valueB\x19\n\x17_uses_internal_standardB\x10\n\x0e_is_normalizedB\x1a\n\x18_uses_authentic_standard\"\x19\n\x08\x44\x61teTime\x12\r\n\x05value\x18\x01 \x01(\t\"\xcc\x04\n\x08\x41nalysis\x12(\n\x04type\x18\x01 \x01(\x0e\x32\x1a.ord.Analysis.AnalysisType\x12\x0f\n\x07\x64\x65tails\x18\x02 \x01(\t\x12\x0f\n\x07\x63hmo_id\x18\x03 \x01(\x05\x12#\n\x16is_of_isolated_species\x18\x04 \x01(\x08H\x00\x88\x01\x01\x12%\n\x04\x64\x61ta\x18\x05 \x03(\x0b\x32\x17.ord.Analysis.DataEntry\x12\x1f\n\x17instrument_manufacturer\x18\x06 \x01(\t\x12\x31\n\x1ainstrument_last_calibrated\x18\x07 \x01(\x0b\x32\r.ord.DateTime\x1a\x36\n\tDataEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x18\n\x05value\x18\x02 \x01(\x0b\x32\t.ord.Data:\x02\x38\x01\"\x80\x02\n\x0c\x41nalysisType\x12\x0f\n\x0bUNSPECIFIED\x10\x00\x12\n\n\x06\x43USTOM\x10\x01\x12\x06\n\x02LC\x10\x02\x12\x06\n\x02GC\x10\x03\x12\x06\n\x02IR\x10\x04\x12\n\n\x06NMR_1H\x10\x05\x12\x0b\n\x07NMR_13C\x10\x06\x12\r\n\tNMR_OTHER\x10\x07\x12\x06\n\x02MP\x10\x08\x12\x06\n\x02UV\x10\t\x12\x07\n\x03TLC\x10\n\x12\x06\n\x02MS\x10\x0b\x12\x08\n\x04HRMS\x10\x0c\x12\x08\n\x04MSMS\x10\r\x12\n\n\x06WEIGHT\x10\x0e\x12\x08\n\x04LCMS\x10\x0f\x12\x08\n\x04GCMS\x10\x10\x12\x08\n\x04\x45LSD\x10\x11\x12\x06\n\x02\x43\x44\x10\x12\x12\x07\n\x03SFC\x10\x13\x12\x07\n\x03\x45PR\x10\x14\x12\x07\n\x03XRD\x10\x15\x12\t\n\x05RAMAN\x10\x16\x12\x06\n\x02\x45\x44\x10\x17\x42\x19\n\x17_is_of_isolated_species\"\x87\x03\n\x12ReactionProvenance\x12!\n\x0c\x65xperimenter\x18\x01 \x01(\x0b\x32\x0b.ord.Person\x12\x0c\n\x04\x63ity\x18\x02 \x01(\t\x12\'\n\x10\x65xperiment_start\x18\x03 \x01(\x0b\x32\r.ord.DateTime\x12\x0b\n\x03\x64oi\x18\x04 \x01(\t\x12\x0e\n\x06patent\x18\x05 \x01(\t\x12\x17\n\x0fpublication_url\x18\x06 \x01(\t\x12(\n\x0erecord_created\x18\x07 \x01(\x0b\x32\x10.ord.RecordEvent\x12)\n\x0frecord_modified\x18\x08 \x03(\x0b\x32\x10.ord.RecordEvent\x12H\n\x11reaction_metadata\x18\t \x03(\x0b\x32-.ord.ReactionProvenance.ReactionMetadataEntry\x1a\x42\n\x15ReactionMetadataEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x18\n\x05value\x18\x02 \x01(\x0b\x32\t.ord.Data:\x02\x38\x01\"\\\n\x06Person\x12\x10\n\x08username\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\r\n\x05orcid\x18\x03 \x01(\t\x12\x14\n\x0corganization\x18\x04 \x01(\t\x12\r\n\x05\x65mail\x18\x05 \x01(\t\"X\n\x0bRecordEvent\x12\x1b\n\x04time\x18\x01 \x01(\x0b\x32\r.ord.DateTime\x12\x1b\n\x06person\x18\x02 \x01(\x0b\x32\x0b.ord.Person\x12\x0f\n\x07\x64\x65tails\x18\x03 \x01(\t\"\xb5\x01\n\x04Time\x12\x12\n\x05value\x18\x01 \x01(\x02H\x00\x88\x01\x01\x12\x16\n\tprecision\x18\x02 \x01(\x02H\x01\x88\x01\x01\x12!\n\x05units\x18\x03 \x01(\x0e\x32\x12.ord.Time.TimeUnit\"F\n\x08TimeUnit\x12\x0f\n\x0bUNSPECIFIED\x10\x00\x12\x07\n\x03\x44\x41Y\x10\x04\x12\x08\n\x04HOUR\x10\x01\x12\n\n\x06MINUTE\x10\x02\x12\n\n\x06SECOND\x10\x03\x42\x08\n\x06_valueB\x0c\n\n_precision\"\xc0\x01\n\x04Mass\x12\x12\n\x05value\x18\x01 \x01(\x02H\x00\x88\x01\x01\x12\x16\n\tprecision\x18\x02 \x01(\x02H\x01\x88\x01\x01\x12!\n\x05units\x18\x03 \x01(\x0e\x32\x12.ord.Mass.MassUnit\"Q\n\x08MassUnit\x12\x0f\n\x0bUNSPECIFIED\x10\x00\x12\x0c\n\x08KILOGRAM\x10\x01\x12\x08\n\x04GRAM\x10\x02\x12\r\n\tMILLIGRAM\x10\x03\x12\r\n\tMICROGRAM\x10\x04\x42\x08\n\x06_valueB\x0c\n\n_precision\"\xc4\x01\n\x05Moles\x12\x12\n\x05value\x18\x01 \x01(\x02H\x00\x88\x01\x01\x12\x16\n\tprecision\x18\x02 \x01(\x02H\x01\x88\x01\x01\x12#\n\x05units\x18\x03 \x01(\x0e\x32\x14.ord.Moles.MolesUnit\"R\n\tMolesUnit\x12\x0f\n\x0bUNSPECIFIED\x10\x00\x12\x08\n\x04MOLE\x10\x01\x12\r\n\tMILLIMOLE\x10\x02\x12\r\n\tMICROMOLE\x10\x03\x12\x0c\n\x08NANOMOLE\x10\x04\x42\x08\n\x06_valueB\x0c\n\n_precision\"\xcc\x01\n\x06Volume\x12\x12\n\x05value\x18\x01 \x01(\x02H\x00\x88\x01\x01\x12\x16\n\tprecision\x18\x02 \x01(\x02H\x01\x88\x01\x01\x12%\n\x05units\x18\x03 \x01(\x0e\x32\x16.ord.Volume.VolumeUnit\"W\n\nVolumeUnit\x12\x0f\n\x0bUNSPECIFIED\x10\x00\x12\t\n\x05LITER\x10\x01\x12\x0e\n\nMILLILITER\x10\x02\x12\x0e\n\nMICROLITER\x10\x03\x12\r\n\tNANOLITER\x10\x04\x42\x08\n\x06_valueB\x0c\n\n_precision\"\xd9\x01\n\rConcentration\x12\x12\n\x05value\x18\x01 \x01(\x02H\x00\x88\x01\x01\x12\x16\n\tprecision\x18\x02 \x01(\x02H\x01\x88\x01\x01\x12\x33\n\x05units\x18\x03 \x01(\x0e\x32$.ord.Concentration.ConcentrationUnit\"O\n\x11\x43oncentrationUnit\x12\x0f\n\x0bUNSPECIFIED\x10\x00\x12\t\n\x05MOLAR\x10\x01\x12\x0e\n\nMILLIMOLAR\x10\x02\x12\x0e\n\nMICROMOLAR\x10\x03\x42\x08\n\x06_valueB\x0c\n\n_precision\"\xf7\x01\n\x08Pressure\x12\x12\n\x05value\x18\x01 \x01(\x02H\x00\x88\x01\x01\x12\x16\n\tprecision\x18\x02 \x01(\x02H\x01\x88\x01\x01\x12)\n\x05units\x18\x03 \x01(\x0e\x32\x1a.ord.Pressure.PressureUnit\"|\n\x0cPressureUnit\x12\x0f\n\x0bUNSPECIFIED\x10\x00\x12\x07\n\x03\x42\x41R\x10\x01\x12\x0e\n\nATMOSPHERE\x10\x02\x12\x07\n\x03PSI\x10\x03\x12\x08\n\x04KPSI\x10\x04\x12\n\n\x06PASCAL\x10\x05\x12\x0e\n\nKILOPASCAL\x10\x06\x12\x08\n\x04TORR\x10\x07\x12\t\n\x05MM_HG\x10\x08\x42\x08\n\x06_valueB\x0c\n\n_precision\"\xcf\x01\n\x0bTemperature\x12\x12\n\x05value\x18\x01 \x01(\x02H\x00\x88\x01\x01\x12\x16\n\tprecision\x18\x02 \x01(\x02H\x01\x88\x01\x01\x12/\n\x05units\x18\x03 \x01(\x0e\x32 .ord.Temperature.TemperatureUnit\"K\n\x0fTemperatureUnit\x12\x0f\n\x0bUNSPECIFIED\x10\x00\x12\x0b\n\x07\x43\x45LSIUS\x10\x01\x12\x0e\n\nFAHRENHEIT\x10\x02\x12\n\n\x06KELVIN\x10\x03\x42\x08\n\x06_valueB\x0c\n\n_precision\"\xb3\x01\n\x07\x43urrent\x12\x12\n\x05value\x18\x01 \x01(\x02H\x00\x88\x01\x01\x12\x16\n\tprecision\x18\x02 \x01(\x02H\x01\x88\x01\x01\x12\'\n\x05units\x18\x03 \x01(\x0e\x32\x18.ord.Current.CurrentUnit\";\n\x0b\x43urrentUnit\x12\x0f\n\x0bUNSPECIFIED\x10\x00\x12\n\n\x06\x41MPERE\x10\x01\x12\x0f\n\x0bMILLIAMPERE\x10\x02\x42\x08\n\x06_valueB\x0c\n\n_precision\"\xaf\x01\n\x07Voltage\x12\x12\n\x05value\x18\x01 \x01(\x02H\x00\x88\x01\x01\x12\x16\n\tprecision\x18\x02 \x01(\x02H\x01\x88\x01\x01\x12\'\n\x05units\x18\x03 \x01(\x0e\x32\x18.ord.Voltage.VoltageUnit\"7\n\x0bVoltageUnit\x12\x0f\n\x0bUNSPECIFIED\x10\x00\x12\x08\n\x04VOLT\x10\x01\x12\r\n\tMILLIVOLT\x10\x02\x42\x08\n\x06_valueB\x0c\n\n_precision\"\xd1\x01\n\x06Length\x12\x12\n\x05value\x18\x01 \x01(\x02H\x00\x88\x01\x01\x12\x16\n\tprecision\x18\x02 \x01(\x02H\x01\x88\x01\x01\x12%\n\x05units\x18\x03 \x01(\x0e\x32\x16.ord.Length.LengthUnit\"\\\n\nLengthUnit\x12\x0f\n\x0bUNSPECIFIED\x10\x00\x12\x0e\n\nCENTIMETER\x10\x01\x12\x0e\n\nMILLIMETER\x10\x02\x12\t\n\x05METER\x10\x03\x12\x08\n\x04INCH\x10\x04\x12\x08\n\x04\x46OOT\x10\x05\x42\x08\n\x06_valueB\x0c\n\n_precision\"\xc1\x01\n\nWavelength\x12\x12\n\x05value\x18\x01 \x01(\x02H\x00\x88\x01\x01\x12\x16\n\tprecision\x18\x02 \x01(\x02H\x01\x88\x01\x01\x12-\n\x05units\x18\x03 \x01(\x0e\x32\x1e.ord.Wavelength.WavelengthUnit\"@\n\x0eWavelengthUnit\x12\x0f\n\x0bUNSPECIFIED\x10\x00\x12\r\n\tNANOMETER\x10\x01\x12\x0e\n\nWAVENUMBER\x10\x02\x42\x08\n\x06_valueB\x0c\n\n_precision\"\xa0\x02\n\x08\x46lowRate\x12\x12\n\x05value\x18\x01 \x01(\x02H\x00\x88\x01\x01\x12\x16\n\tprecision\x18\x02 \x01(\x02H\x01\x88\x01\x01\x12)\n\x05units\x18\x03 \x01(\x0e\x32\x1a.ord.FlowRate.FlowRateUnit\"\xa4\x01\n\x0c\x46lowRateUnit\x12\x0f\n\x0bUNSPECIFIED\x10\x00\x12\x19\n\x15MICROLITER_PER_MINUTE\x10\x01\x12\x19\n\x15MICROLITER_PER_SECOND\x10\x02\x12\x19\n\x15MILLILITER_PER_MINUTE\x10\x03\x12\x19\n\x15MILLILITER_PER_SECOND\x10\x04\x12\x17\n\x13MICROLITER_PER_HOUR\x10\x05\x42\x08\n\x06_valueB\x0c\n\n_precision\"P\n\nPercentage\x12\x12\n\x05value\x18\x01 \x01(\x02H\x00\x88\x01\x01\x12\x16\n\tprecision\x18\x02 \x01(\x02H\x01\x88\x01\x01\x42\x08\n\x06_valueB\x0c\n\n_precision\"P\n\nFloatValue\x12\x12\n\x05value\x18\x01 \x01(\x02H\x00\x88\x01\x01\x12\x16\n\tprecision\x18\x02 \x01(\x02H\x01\x88\x01\x01\x42\x08\n\x06_valueB\x0c\n\n_precision\"\xa1\x01\n\x04\x44\x61ta\x12\x15\n\x0b\x66loat_value\x18\x01 \x01(\x02H\x00\x12\x17\n\rinteger_value\x18\x02 \x01(\x05H\x00\x12\x15\n\x0b\x62ytes_value\x18\x03 \x01(\x0cH\x00\x12\x16\n\x0cstring_value\x18\x04 \x01(\tH\x00\x12\r\n\x03url\x18\x05 \x01(\tH\x00\x12\x13\n\x0b\x64\x65scription\x18\x06 \x01(\t\x12\x0e\n\x06\x66ormat\x18\x07 \x01(\tB\x06\n\x04kindb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'ord_schema.proto.reaction_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _REACTION_INPUTSENTRY._options = None
  _REACTION_INPUTSENTRY._serialized_options = b'8\001'
  _COMPOUND_FEATURESENTRY._options = None
  _COMPOUND_FEATURESENTRY._serialized_options = b'8\001'
  _COMPOUND_ANALYSESENTRY._options = None
  _COMPOUND_ANALYSESENTRY._serialized_options = b'8\001'
  _REACTIONSETUP_AUTOMATIONCODEENTRY._options = None
  _REACTIONSETUP_AUTOMATIONCODEENTRY._serialized_options = b'8\001'
  _REACTIONOUTCOME_ANALYSESENTRY._options = None
  _REACTIONOUTCOME_ANALYSESENTRY._serialized_options = b'8\001'
  _PRODUCTCOMPOUND_FEATURESENTRY._options = None
  _PRODUCTCOMPOUND_FEATURESENTRY._serialized_options = b'8\001'
  _ANALYSIS_DATAENTRY._options = None
  _ANALYSIS_DATAENTRY._serialized_options = b'8\001'
  _REACTIONPROVENANCE_REACTIONMETADATAENTRY._options = None
  _REACTIONPROVENANCE_REACTIONMETADATAENTRY._serialized_options = b'8\001'
  _globals['_REACTION']._serialized_start=41
  _globals['_REACTION']._serialized_end=514
  _globals['_REACTION_INPUTSENTRY']._serialized_start=449
  _globals['_REACTION_INPUTSENTRY']._serialized_end=514
  _globals['_REACTIONIDENTIFIER']._serialized_start=517
  _globals['_REACTIONIDENTIFIER']._serialized_end=812
  _globals['_REACTIONIDENTIFIER_REACTIONIDENTIFIERTYPE']._serialized_start=658
  _globals['_REACTIONIDENTIFIER_REACTIONIDENTIFIERTYPE']._serialized_end=798
  _globals['_REACTIONINPUT']._serialized_start=815
  _globals['_REACTIONINPUT']._serialized_end=1643
  _globals['_REACTIONINPUT_ADDITIONSPEED']._serialized_start=1211
  _globals['_REACTIONINPUT_ADDITIONSPEED']._serialized_end=1431
  _globals['_REACTIONINPUT_ADDITIONSPEED_ADDITIONSPEEDTYPE']._serialized_start=1311
  _globals['_REACTIONINPUT_ADDITIONSPEED_ADDITIONSPEEDTYPE']._serialized_end=1431
  _globals['_REACTIONINPUT_ADDITIONDEVICE']._serialized_start=1434
  _globals['_REACTIONINPUT_ADDITIONDEVICE']._serialized_end=1643
  _globals['_REACTIONINPUT_ADDITIONDEVICE_ADDITIONDEVICETYPE']._serialized_start=1537
  _globals['_REACTIONINPUT_ADDITIONDEVICE_ADDITIONDEVICETYPE']._serialized_end=1643
  _globals['_AMOUNT']._serialized_start=1646
  _globals['_AMOUNT']._serialized_end=1860
  _globals['_UNMEASUREDAMOUNT']._serialized_start=1863
  _globals['_UNMEASUREDAMOUNT']._serialized_end=2053
  _globals['_UNMEASUREDAMOUNT_UNMEASUREDAMOUNTTYPE']._serialized_start=1958
  _globals['_UNMEASUREDAMOUNT_UNMEASUREDAMOUNTTYPE']._serialized_end=2053
  _globals['_CRUDECOMPONENT']._serialized_start=2056
  _globals['_CRUDECOMPONENT']._serialized_end=2228
  _globals['_COMPOUND']._serialized_start=2231
  _globals['_COMPOUND']._serialized_end=2780
  _globals['_COMPOUND_SOURCE']._serialized_start=2583
  _globals['_COMPOUND_SOURCE']._serialized_end=2640
  _globals['_COMPOUND_FEATURESENTRY']._serialized_start=2642
  _globals['_COMPOUND_FEATURESENTRY']._serialized_end=2700
  _globals['_COMPOUND_ANALYSESENTRY']._serialized_start=2702
  _globals['_COMPOUND_ANALYSESENTRY']._serialized_end=2764
  _globals['_REACTIONROLE']._serialized_start=2783
  _globals['_REACTIONROLE']._serialized_end=2961
  _globals['_REACTIONROLE_REACTIONROLETYPE']._serialized_start=2800
  _globals['_REACTIONROLE_REACTIONROLETYPE']._serialized_end=2961
  _globals['_COMPOUNDPREPARATION']._serialized_start=2964
  _globals['_COMPOUNDPREPARATION']._serialized_end=3210
  _globals['_COMPOUNDPREPARATION_COMPOUNDPREPARATIONTYPE']._serialized_start=3089
  _globals['_COMPOUNDPREPARATION_COMPOUNDPREPARATIONTYPE']._serialized_end=3210
  _globals['_COMPOUNDIDENTIFIER']._serialized_start=3213
  _globals['_COMPOUNDIDENTIFIER']._serialized_end=3599
  _globals['_COMPOUNDIDENTIFIER_COMPOUNDIDENTIFIERTYPE']._serialized_start=3330
  _globals['_COMPOUNDIDENTIFIER_COMPOUNDIDENTIFIERTYPE']._serialized_end=3599
  _globals['_VESSEL']._serialized_start=3602
  _globals['_VESSEL']._serialized_end=4153
  _globals['_VESSEL_VESSELTYPE']._serialized_start=3889
  _globals['_VESSEL_VESSELTYPE']._serialized_end=4153
  _globals['_VESSELMATERIAL']._serialized_start=4156
  _globals['_VESSELMATERIAL']._serialized_end=4360
  _globals['_VESSELMATERIAL_VESSELMATERIALTYPE']._serialized_start=4245
  _globals['_VESSELMATERIAL_VESSELMATERIALTYPE']._serialized_end=4360
  _globals['_VESSELATTACHMENT']._serialized_start=4363
  _globals['_VESSELATTACHMENT']._serialized_end=4770
  _globals['_VESSELATTACHMENT_VESSELATTACHMENTTYPE']._serialized_start=4459
  _globals['_VESSELATTACHMENT_VESSELATTACHMENTTYPE']._serialized_end=4770
  _globals['_VESSELPREPARATION']._serialized_start=4773
  _globals['_VESSELPREPARATION']._serialized_end=5005
  _globals['_VESSELPREPARATION_VESSELPREPARATIONTYPE']._serialized_start=4872
  _globals['_VESSELPREPARATION_VESSELPREPARATIONTYPE']._serialized_end=5005
  _globals['_REACTIONSETUP']._serialized_start=5008
  _globals['_REACTIONSETUP']._serialized_end=5552
  _globals['_REACTIONSETUP_AUTOMATIONCODEENTRY']._serialized_start=5236
  _globals['_REACTIONSETUP_AUTOMATIONCODEENTRY']._serialized_end=5300
  _globals['_REACTIONSETUP_REACTIONENVIRONMENT']._serialized_start=5303
  _globals['_REACTIONSETUP_REACTIONENVIRONMENT']._serialized_end=5535
  _globals['_REACTIONSETUP_REACTIONENVIRONMENT_REACTIONENVIRONMENTTYPE']._serialized_start=5421
  _globals['_REACTIONSETUP_REACTIONENVIRONMENT_REACTIONENVIRONMENTTYPE']._serialized_end=5535
  _globals['_REACTIONCONDITIONS']._serialized_start=5555
  _globals['_REACTIONCONDITIONS']._serialized_end=5992
  _globals['_TEMPERATURECONDITIONS']._serialized_start=5995
  _globals['_TEMPERATURECONDITIONS']._serialized_end=6861
  _globals['_TEMPERATURECONDITIONS_TEMPERATURECONTROL']._serialized_start=6194
  _globals['_TEMPERATURECONDITIONS_TEMPERATURECONTROL']._serialized_end=6534
  _globals['_TEMPERATURECONDITIONS_TEMPERATURECONTROL_TEMPERATURECONTROLTYPE']._serialized_start=6318
  _globals['_TEMPERATURECONDITIONS_TEMPERATURECONTROL_TEMPERATURECONTROLTYPE']._serialized_end=6534
  _globals['_TEMPERATURECONDITIONS_TEMPERATUREMEASUREMENT']._serialized_start=6537
  _globals['_TEMPERATURECONDITIONS_TEMPERATUREMEASUREMENT']._serialized_end=6861
  _globals['_TEMPERATURECONDITIONS_TEMPERATUREMEASUREMENT_TEMPERATUREMEASUREMENTTYPE']._serialized_start=6736
  _globals['_TEMPERATURECONDITIONS_TEMPERATUREMEASUREMENT_TEMPERATUREMEASUREMENTTYPE']._serialized_end=6861
  _globals['_PRESSURECONDITIONS']._serialized_start=6864
  _globals['_PRESSURECONDITIONS']._serialized_end=7900
  _globals['_PRESSURECONDITIONS_PRESSURECONTROL']._serialized_start=7101
  _globals['_PRESSURECONDITIONS_PRESSURECONTROL']._serialized_end=7325
  _globals['_PRESSURECONDITIONS_PRESSURECONTROL_PRESSURECONTROLTYPE']._serialized_start=7212
  _globals['_PRESSURECONDITIONS_PRESSURECONTROL_PRESSURECONTROLTYPE']._serialized_end=7325
  _globals['_PRESSURECONDITIONS_ATMOSPHERE']._serialized_start=7328
  _globals['_PRESSURECONDITIONS_ATMOSPHERE']._serialized_end=7637
  _globals['_PRESSURECONDITIONS_ATMOSPHERE_ATMOSPHERETYPE']._serialized_start=7425
  _globals['_PRESSURECONDITIONS_ATMOSPHERE_ATMOSPHERETYPE']._serialized_end=7637
  _globals['_PRESSURECONDITIONS_PRESSUREMEASUREMENT']._serialized_start=7640
  _globals['_PRESSURECONDITIONS_PRESSUREMEASUREMENT']._serialized_end=7900
  _globals['_PRESSURECONDITIONS_PRESSUREMEASUREMENT_PRESSUREMEASUREMENTTYPE']._serialized_start=7821
  _globals['_PRESSURECONDITIONS_PRESSUREMEASUREMENT_PRESSUREMEASUREMENTTYPE']._serialized_end=7900
  _globals['_STIRRINGCONDITIONS']._serialized_start=7903
  _globals['_STIRRINGCONDITIONS']._serialized_end=8379
  _globals['_STIRRINGCONDITIONS_STIRRINGRATE']._serialized_start=8053
  _globals['_STIRRINGCONDITIONS_STIRRINGRATE']._serialized_end=8234
  _globals['_STIRRINGCONDITIONS_STIRRINGRATE_STIRRINGRATETYPE']._serialized_start=8168
  _globals['_STIRRINGCONDITIONS_STIRRINGRATE_STIRRINGRATETYPE']._serialized_end=8234
  _globals['_STIRRINGCONDITIONS_STIRRINGMETHODTYPE']._serialized_start=8237
  _globals['_STIRRINGCONDITIONS_STIRRINGMETHODTYPE']._serialized_end=8379
  _globals['_ILLUMINATIONCONDITIONS']._serialized_start=8382
  _globals['_ILLUMINATIONCONDITIONS']._serialized_end=8742
  _globals['_ILLUMINATIONCONDITIONS_ILLUMINATIONTYPE']._serialized_start=8584
  _globals['_ILLUMINATIONCONDITIONS_ILLUMINATIONTYPE']._serialized_end=8742
  _globals['_ELECTROCHEMISTRYCONDITIONS']._serialized_start=8745
  _globals['_ELECTROCHEMISTRYCONDITIONS']._serialized_end=9609
  _globals['_ELECTROCHEMISTRYCONDITIONS_ELECTROCHEMISTRYMEASUREMENT']._serialized_start=9166
  _globals['_ELECTROCHEMISTRYCONDITIONS_ELECTROCHEMISTRYMEASUREMENT']._serialized_end=9282
  _globals['_ELECTROCHEMISTRYCONDITIONS_ELECTROCHEMISTRYCELL']._serialized_start=9285
  _globals['_ELECTROCHEMISTRYCONDITIONS_ELECTROCHEMISTRYCELL']._serialized_end=9512
  _globals['_ELECTROCHEMISTRYCONDITIONS_ELECTROCHEMISTRYCELL_ELECTROCHEMISTRYCELLTYPE']._serialized_start=9419
  _globals['_ELECTROCHEMISTRYCONDITIONS_ELECTROCHEMISTRYCELL_ELECTROCHEMISTRYCELLTYPE']._serialized_end=9512
  _globals['_ELECTROCHEMISTRYCONDITIONS_ELECTROCHEMISTRYTYPE']._serialized_start=9514
  _globals['_ELECTROCHEMISTRYCONDITIONS_ELECTROCHEMISTRYTYPE']._serialized_end=9609
  _globals['_FLOWCONDITIONS']._serialized_start=9612
  _globals['_FLOWCONDITIONS']._serialized_end=10144
  _globals['_FLOWCONDITIONS_TUBING']._serialized_start=9755
  _globals['_FLOWCONDITIONS_TUBING']._serialized_end=10019
  _globals['_FLOWCONDITIONS_TUBING_TUBINGTYPE']._serialized_start=9867
  _globals['_FLOWCONDITIONS_TUBING_TUBINGTYPE']._serialized_end=10019
  _globals['_FLOWCONDITIONS_FLOWTYPE']._serialized_start=10021
  _globals['_FLOWCONDITIONS_FLOWTYPE']._serialized_end=10144
  _globals['_REACTIONNOTES']._serialized_start=10147
  _globals['_REACTIONNOTES']._serialized_end=10595
  _globals['_REACTIONOBSERVATION']._serialized_start=10597
  _globals['_REACTIONOBSERVATION']._serialized_end=10686
  _globals['_REACTIONWORKUP']._serialized_start=10689
  _globals['_REACTIONWORKUP']._serialized_end=11404
  _globals['_REACTIONWORKUP_REACTIONWORKUPTYPE']._serialized_start=11035
  _globals['_REACTIONWORKUP_REACTIONWORKUPTYPE']._serialized_end=11373
  _globals['_REACTIONOUTCOME']._serialized_start=11407
  _globals['_REACTIONOUTCOME']._serialized_end=11653
  _globals['_REACTIONOUTCOME_ANALYSESENTRY']._serialized_start=2702
  _globals['_REACTIONOUTCOME_ANALYSESENTRY']._serialized_end=2764
  _globals['_PRODUCTCOMPOUND']._serialized_start=11656
  _globals['_PRODUCTCOMPOUND']._serialized_end=12309
  _globals['_PRODUCTCOMPOUND_TEXTURE']._serialized_start=11986
  _globals['_PRODUCTCOMPOUND_TEXTURE']._serialized_end=12226
  _globals['_PRODUCTCOMPOUND_TEXTURE_TEXTURETYPE']._serialized_start=12071
  _globals['_PRODUCTCOMPOUND_TEXTURE_TEXTURETYPE']._serialized_end=12226
  _globals['_PRODUCTCOMPOUND_FEATURESENTRY']._serialized_start=2642
  _globals['_PRODUCTCOMPOUND_FEATURESENTRY']._serialized_end=2700
  _globals['_PRODUCTMEASUREMENT']._serialized_start=12312
  _globals['_PRODUCTMEASUREMENT']._serialized_end=13715
  _globals['_PRODUCTMEASUREMENT_MASSSPECMEASUREMENTDETAILS']._serialized_start=12925
  _globals['_PRODUCTMEASUREMENT_MASSSPECMEASUREMENTDETAILS']._serialized_end=13286
  _globals['_PRODUCTMEASUREMENT_MASSSPECMEASUREMENTDETAILS_MASSSPECMEASUREMENTTYPE']._serialized_start=13140
  _globals['_PRODUCTMEASUREMENT_MASSSPECMEASUREMENTDETAILS_MASSSPECMEASUREMENTTYPE']._serialized_end=13248
  _globals['_PRODUCTMEASUREMENT_SELECTIVITY']._serialized_start=13289
  _globals['_PRODUCTMEASUREMENT_SELECTIVITY']._serialized_end=13474
  _globals['_PRODUCTMEASUREMENT_SELECTIVITY_SELECTIVITYTYPE']._serialized_start=13388
  _globals['_PRODUCTMEASUREMENT_SELECTIVITY_SELECTIVITYTYPE']._serialized_end=13474
  _globals['_PRODUCTMEASUREMENT_PRODUCTMEASUREMENTTYPE']._serialized_start=13477
  _globals['_PRODUCTMEASUREMENT_PRODUCTMEASUREMENTTYPE']._serialized_end=13633
  _globals['_DATETIME']._serialized_start=13717
  _globals['_DATETIME']._serialized_end=13742
  _globals['_ANALYSIS']._serialized_start=13745
  _globals['_ANALYSIS']._serialized_end=14333
  _globals['_ANALYSIS_DATAENTRY']._serialized_start=13993
  _globals['_ANALYSIS_DATAENTRY']._serialized_end=14047
  _globals['_ANALYSIS_ANALYSISTYPE']._serialized_start=14050
  _globals['_ANALYSIS_ANALYSISTYPE']._serialized_end=14306
  _globals['_REACTIONPROVENANCE']._serialized_start=14336
  _globals['_REACTIONPROVENANCE']._serialized_end=14727
  _globals['_REACTIONPROVENANCE_REACTIONMETADATAENTRY']._serialized_start=14661
  _globals['_REACTIONPROVENANCE_REACTIONMETADATAENTRY']._serialized_end=14727
  _globals['_PERSON']._serialized_start=14729
  _globals['_PERSON']._serialized_end=14821
  _globals['_RECORDEVENT']._serialized_start=14823
  _globals['_RECORDEVENT']._serialized_end=14911
  _globals['_TIME']._serialized_start=14914
  _globals['_TIME']._serialized_end=15095
  _globals['_TIME_TIMEUNIT']._serialized_start=15001
  _globals['_TIME_TIMEUNIT']._serialized_end=15071
  _globals['_MASS']._serialized_start=15098
  _globals['_MASS']._serialized_end=15290
  _globals['_MASS_MASSUNIT']._serialized_start=15185
  _globals['_MASS_MASSUNIT']._serialized_end=15266
  _globals['_MOLES']._serialized_start=15293
  _globals['_MOLES']._serialized_end=15489
  _globals['_MOLES_MOLESUNIT']._serialized_start=15383
  _globals['_MOLES_MOLESUNIT']._serialized_end=15465
  _globals['_VOLUME']._serialized_start=15492
  _globals['_VOLUME']._serialized_end=15696
  _globals['_VOLUME_VOLUMEUNIT']._serialized_start=15585
  _globals['_VOLUME_VOLUMEUNIT']._serialized_end=15672
  _globals['_CONCENTRATION']._serialized_start=15699
  _globals['_CONCENTRATION']._serialized_end=15916
  _globals['_CONCENTRATION_CONCENTRATIONUNIT']._serialized_start=15813
  _globals['_CONCENTRATION_CONCENTRATIONUNIT']._serialized_end=15892
  _globals['_PRESSURE']._serialized_start=15919
  _globals['_PRESSURE']._serialized_end=16166
  _globals['_PRESSURE_PRESSUREUNIT']._serialized_start=16018
  _globals['_PRESSURE_PRESSUREUNIT']._serialized_end=16142
  _globals['_TEMPERATURE']._serialized_start=16169
  _globals['_TEMPERATURE']._serialized_end=16376
  _globals['_TEMPERATURE_TEMPERATUREUNIT']._serialized_start=16277
  _globals['_TEMPERATURE_TEMPERATUREUNIT']._serialized_end=16352
  _globals['_CURRENT']._serialized_start=16379
  _globals['_CURRENT']._serialized_end=16558
  _globals['_CURRENT_CURRENTUNIT']._serialized_start=16475
  _globals['_CURRENT_CURRENTUNIT']._serialized_end=16534
  _globals['_VOLTAGE']._serialized_start=16561
  _globals['_VOLTAGE']._serialized_end=16736
  _globals['_VOLTAGE_VOLTAGEUNIT']._serialized_start=16657
  _globals['_VOLTAGE_VOLTAGEUNIT']._serialized_end=16712
  _globals['_LENGTH']._serialized_start=16739
  _globals['_LENGTH']._serialized_end=16948
  _globals['_LENGTH_LENGTHUNIT']._serialized_start=16832
  _globals['_LENGTH_LENGTHUNIT']._serialized_end=16924
  _globals['_WAVELENGTH']._serialized_start=16951
  _globals['_WAVELENGTH']._serialized_end=17144
  _globals['_WAVELENGTH_WAVELENGTHUNIT']._serialized_start=17056
  _globals['_WAVELENGTH_WAVELENGTHUNIT']._serialized_end=17120
  _globals['_FLOWRATE']._serialized_start=17147
  _globals['_FLOWRATE']._serialized_end=17435
  _globals['_FLOWRATE_FLOWRATEUNIT']._serialized_start=17247
  _globals['_FLOWRATE_FLOWRATEUNIT']._serialized_end=17411
  _globals['_PERCENTAGE']._serialized_start=17437
  _globals['_PERCENTAGE']._serialized_end=17517
  _globals['_FLOATVALUE']._serialized_start=17519
  _globals['_FLOATVALUE']._serialized_end=17599
  _globals['_DATA']._serialized_start=17602
  _globals['_DATA']._serialized_end=17763
# @@protoc_insertion_point(module_scope)



# Copyright 2022 Open Reaction Database Project Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""RDKit PostgreSQL cartridge functionality.

Notes:
  * These tables live in a separate "rdkit" schema to avoid name conflicts between tables and extension types.
  * The RDKit-specific columns are populated by ord_schema.orm.database.add_rdkit; this allows the ORM to function
    normally even if if the RDKit PostgreSQL cartridge is not installed (the `smiles` column will be populated and
    the other columns will be empty).
  * Objects with this type are added to the ORM in from_proto() using the `rdkit` field.
"""
from __future__ import annotations

import os
from distutils.util import strtobool  # pylint: disable=deprecated-module
from enum import Enum

from sqlalchemy import Column, Index, Integer, ForeignKey, Text, func
from sqlalchemy.types import UserDefinedType

from ord_schema.orm import Base


def rdkit_cartridge() -> bool:
    """Returns whether to use RDKit PostgreSQL cartridge functionality."""
    return bool(strtobool(os.environ.get("ORD_POSTGRES_RDKIT", "1")))


class _RDKitMol(UserDefinedType):
    """https://github.com/rdkit/rdkit/blob/master/Code/PgSQL/rdkit/rdkit.sql.in#L4."""

    cache_ok = True

    @property
    def python_type(self):
        raise NotImplementedError

    def get_col_spec(self, **kwargs):
        """Returns the column type."""
        del kwargs  # Unused.
        return "rdkit.mol" if rdkit_cartridge() else "bytea"


class _RDKitReaction(UserDefinedType):
    """https://github.com/rdkit/rdkit/blob/master/Code/PgSQL/rdkit/rdkit.sql.in#L129."""

    cache_ok = True

    @property
    def python_type(self):
        raise NotImplementedError

    def get_col_spec(self, **kwargs):
        """Returns the column type."""
        del kwargs  # Unused.
        return "rdkit.reaction" if rdkit_cartridge() else "bytea"


class _RDKitBfp(UserDefinedType):
    """https://github.com/rdkit/rdkit/blob/master/Code/PgSQL/rdkit/rdkit.sql.in#L81."""

    cache_ok = True

    @property
    def python_type(self):
        raise NotImplementedError

    def get_col_spec(self, **kwargs):
        """Returns the column type."""
        del kwargs  # Unused.
        return "rdkit.bfp" if rdkit_cartridge() else "bytea"

    class comparator_factory(  # pylint: disable=abstract-method,invalid-name
        UserDefinedType.Comparator  # pytype: disable=attribute-error
    ):
        def __mod__(self, other):
            return self.bool_op("operator(rdkit.%)")(other)


class _RDKitSfp(UserDefinedType):
    """https://github.com/rdkit/rdkit/blob/master/Code/PgSQL/rdkit/rdkit.sql.in#L105."""

    cache_ok = True

    @property
    def python_type(self):
        raise NotImplementedError

    def get_col_spec(self, **kwargs):
        """Returns the column type."""
        del kwargs  # Unused.
        return "rdkit.sfp" if rdkit_cartridge() else "bytea"

    class comparator_factory(  # pylint: disable=abstract-method,invalid-name
        UserDefinedType.Comparator  # pytype: disable=attribute-error
    ):
        def __mod__(self, other):
            return self.bool_op("operator(rdkit.%)")(other)


class CString(UserDefinedType):
    """PostgreSQL cstring."""

    cache_ok = True

    @property
    def python_type(self):
        raise NotImplementedError

    def get_col_spec(self, **kwargs):
        """Returns the column type."""
        del kwargs  # Unused.
        return "cstring"


class FingerprintType(Enum):
    """RDKit PostgreSQL fingerprint types."""

    MORGAN_BFP = func.rdkit.morganbv_fp
    MORGAN_SFP = func.rdkit.morgan_fp

    def __call__(self, *args, **kwargs):
        return self.value(*args, **kwargs)

    @classmethod
    def get_table_args(cls) -> list:
        """Returns a list of __table_args__ for _Structure.

        Each fingerprint type is given a column (name.lower()) and a corresponding index.

        Returns:
            List of Column and Index objects.
        """
        table_args = []
        for fp_type in cls:
            name = fp_type.name.lower()
            if name.endswith("_bfp"):
                dtype = _RDKitBfp
            elif name.endswith("_sfp"):
                dtype = _RDKitSfp
            else:
                raise ValueError(f"unable to determine dtype for {name}")
            table_args.extend([Column(name, dtype), Index(f"{name}_index", name, postgresql_using="gist")])
        return table_args


class RDKitMol(Base):
    """Table for storing compound structures and associated RDKit cartridge data."""

    __tablename__ = "mols"
    id = Column(Integer, primary_key=True)
    smiles = Column(Text)
    mol = Column(_RDKitMol)

    __table_args__ = (
        Index("mol_index", "mol", postgresql_using="gist"),
        *FingerprintType.get_table_args(),
        {"schema": "rdkit"},
    )

    ord_schema_context = Column(Text, nullable=False)
    __mapper_args__ = {
        "polymorphic_on": ord_schema_context,
        "polymorphic_identity": "mols",
        "with_polymorphic": "*",
    }

    @classmethod
    def tanimoto(cls, other: str, fp_type: FingerprintType = FingerprintType.MORGAN_BFP):
        return func.rdkit.tanimoto_sml(getattr(cls, fp_type.name.lower()), fp_type(other))


class _CompoundRDKit(RDKitMol):
    compound_id = Column(Integer, ForeignKey("compound.id", ondelete="CASCADE"))

    __mapper_args__ = {
        "polymorphic_identity": "Compound.rdkit",
    }


class _ProductCompoundRDKit(RDKitMol):
    product_compound_id = Column(Integer, ForeignKey("product_compound.id", ondelete="CASCADE"))

    __mapper_args__ = {
        "polymorphic_identity": "ProductCompound.rdkit",
    }


class RDKitReaction(Base):
    """Table for storing reaction objects and associated RDKit cartridge data."""

    __tablename__ = "reactions"
    id = Column(Integer, primary_key=True)
    reaction_smiles = Column(Text)
    reaction = Column(_RDKitReaction)

    __table_args__ = (
        Index("reaction_index", "reaction", postgresql_using="gist"),
        {"schema": "rdkit"},
    )

    ord_schema_context = Column(Text, nullable=False)
    __mapper_args__ = {
        "polymorphic_on": ord_schema_context,
        "polymorphic_identity": "reactions",
        "with_polymorphic": "*",
    }


class _ReactionRDKit(RDKitReaction):
    reaction_id = Column(Integer, ForeignKey("reaction.id", ondelete="CASCADE"))

    __mapper_args__ = {
        "polymorphic_identity": "Reaction.rdkit",
    }



# Copyright 2022 Open Reaction Database Project Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Adds datasets to the ORM database.

Usage:
    add_datasets.py --pattern=<str> [options]
    add_datasets.py -h | --help

Options:
    --pattern=<str>         Pattern for dataset filenames
    --overwrite             Overwrite existing datasets
    --url=<str>             Postgres connection string
    --database=<str>        Database [default: orm]
    --username=<str>        Database username [default: postgres]
    --password=<str>        Database password
    --host=<str>            Database host [default: localhost]
    --port=<int>            Database port [default: 5432]
    --n_jobs=<int>          Number of parallel workers [default: 1]
"""
import os
import time
from functools import partial
from glob import glob
from concurrent.futures import ProcessPoolExecutor

from docopt import docopt
from sqlalchemy import create_engine
from sqlalchemy.orm import Session

from ord_schema.logging import get_logger
from ord_schema.message_helpers import load_message
from ord_schema.orm.database import add_dataset, add_rdkit, delete_dataset, get_connection_string
from ord_schema.proto import dataset_pb2

logger = get_logger(__name__)


def _add_dataset(filename: str, url: str, overwrite: bool) -> None:
    """Adds a single dataset to the database.

    Args:
        filename: Dataset filename.
        url: Database connection string.
        overwrite: If True, overwrite an existing dataset.
    """
    logger.info(f"Loading {filename}")
    start = time.time()
    dataset = load_message(filename, dataset_pb2.Dataset)
    logger.info(f"load_message() took {time.time() - start}s")
    engine = create_engine(url, future=True)
    with Session(engine) as session:
        if overwrite:
            delete_dataset(dataset.dataset_id, session)
        add_dataset(dataset, session)
        start = time.time()
        session.commit()
        logger.info(f"session.commit() took {time.time() - start}s")


def main(**kwargs):
    if kwargs.get("--url"):
        url = kwargs["--url"]
    else:
        url = get_connection_string(
            database=kwargs["--database"],
            username=kwargs["--username"],
            password=kwargs["--password"] or os.environ["PGPASSWORD"],
            host=kwargs["--host"],
            port=int(kwargs["--port"]),
        )
    function = partial(_add_dataset, url=url, overwrite=kwargs["--overwrite"])
    filenames = glob(kwargs["--pattern"])
    with ProcessPoolExecutor(max_workers=int(kwargs["--n_jobs"])) as executor:
        for _ in executor.map(function, filenames):
            pass  # Must iterate over results to raise exceptions.
    engine = create_engine(url, future=True)
    with Session(engine) as session:
        add_rdkit(session)
        start = time.time()
        session.commit()
        logger.info(f"session.commit() took {time.time() - start}s")


if __name__ == "__main__":
    main(**docopt(__doc__))



# Copyright 2020 Open Reaction Database Project Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for ord_schema.macros.solutions."""
import itertools

import pytest
from google.protobuf import text_format

from ord_schema.proto import reaction_pb2
from ord_schema.macros import solutions
from ord_schema import validations


def test_simple_solution():
    """Simple input/output pair test."""
    solvent_pbtxt = """
        identifiers {
            type: SMILES
            value: "O"
        }
        amount {
            volume {
                value: 1.0
                units: LITER
            }
            volume_includes_solutes: true
        }
    """
    solute_pbtxt = """
        identifiers {
            type: SMILES
            value: "[Na+].[Cl-]"
        }
        amount {
            moles {
                value: 1.0
                units: MOLE
            }
        }
    """
    expected_solvent = reaction_pb2.Compound()
    expected_solute = reaction_pb2.Compound()
    text_format.Parse(solvent_pbtxt, expected_solvent)
    text_format.Parse(solute_pbtxt, expected_solute)
    actual_compounds = solutions.simple_solution(
        solvent_smiles="O",
        solute_smiles="[Na+].[Cl-]",
        concentration="1M",
        volume="1L",
    )
    assert actual_compounds == [expected_solvent, expected_solute]


@pytest.mark.parametrize(
    "solute_smiles,concentration,volume", itertools.product(["[Na+].[Cl-]", None], ["1M", None], ["1L", None])
)
def test_simple_solution_legal(solute_smiles, concentration, volume):
    """Test that the macro never generates illegal solution configurations."""
    solution_components = solutions.simple_solution(
        solvent_smiles="O",
        solute_smiles=solute_smiles,
        concentration=concentration,
        volume=volume,
    )
    for component in solution_components:
        validations.validate_message(component)


@pytest.mark.parametrize("volume", ["1L", None])
def test_simple_saturated_solution(volume):
    """Test that the macro never generates illegal solution configurations."""
    solution_components = solutions.simple_solution(
        solvent_smiles="O",
        solute_smiles="[Na+].[Cl-]",
        volume=volume,
        saturated=True,
    )
    for component in solution_components:
        validations.validate_message(component)
    assert any(
        component.amount.unmeasured.type == reaction_pb2.UnmeasuredAmount.SATURATED for component in solution_components
    )


def test_simple_solution_concentration_and_saturated_illegal():
    with pytest.raises(ValueError, match="Cannot specify both"):
        solutions.simple_solution(
            solvent_smiles="O",
            solute_smiles="[Na+].[Cl-]",
            volume="1L",
            concentration="1M",
            saturated=True,
        )


def test_simple_solution_saturated_solute_missing():
    with pytest.raises(ValueError, match="Must specify"):
        solutions.simple_solution(solvent_smiles="O", solute_smiles=None, volume="1L", saturated=True)



# Copyright 2022 Open Reaction Database Project Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Table mappings for Reaction protos.

Notes:
    * Foreign keys to the `reaction` table are done using the `id` column, not the ORD reaction ID (`reaction_id`).
      However, the `reaction_id` column is used when the reaction ID is specifically called for, as with crude inputs.
    * We use inheritance to handle messages that appear in more than one context; see
      https://docs.sqlalchemy.org/en/14/orm/inheritance.html. The possible constraints are:
        - Some message types are used more than once in a parent (such as Time in ReactionInput), which forces
          the use of either joined or single table inheritance.
        - Some messages can be repeated, while others are unique to their parent. These uniqueness constraints force
          the use of joined table inheritance since they are specific to their polymorphic type.
      For convenience and consistency, we use single table inheritance for *all* message types, regardless of whether
      they are used in one context or more than one context. This means that we do not enforce the second constraint in
      the database.
"""
from collections import defaultdict
from operator import attrgetter
from typing import Any, Mapping, Optional, Type

from google.protobuf.descriptor import Descriptor, FieldDescriptor
from google.protobuf.message import Message
from inflection import underscore
from sqlalchemy import Boolean, Column, Enum, Float, Integer, ForeignKey, LargeBinary, Text
from sqlalchemy.dialects.postgresql import ARRAY
from sqlalchemy.orm import relationship

import ord_schema.orm.rdkit_mappers  # pylint: disable=unused-import
from ord_schema import message_helpers
from ord_schema.logging import get_logger
from ord_schema.orm import Base
from ord_schema.proto import dataset_pb2
from ord_schema.proto import reaction_pb2


logger = get_logger(__name__)


def get_message_type(full_name: str) -> Any:
    """Fetches the class for a protocol buffer message type."""
    if not full_name.startswith("ord."):
        raise ValueError(f"Unexpected message type: {full_name}")
    operator = attrgetter(full_name[4:])
    try:
        return operator(reaction_pb2)
    except AttributeError:
        return operator(dataset_pb2)


def get_parents(message_type: Type[Message]) -> dict[Type[Message], list[tuple[Type[Message], str, bool]]]:
    """Returns the parent message types for each message type."""
    parents = defaultdict(list)
    for child, parent, field_name, unique in _get_message_contexts(message_type.DESCRIPTOR, None, None, None):
        if parent is not None:
            parents[get_message_type(child)].append((get_message_type(parent), field_name, unique))
    parents[message_type] = []  # Add the base type.
    return parents


def _get_message_contexts(
    descriptor: Descriptor, parent: str | None, field_name: str | None, unique: bool | None
) -> set[tuple[str, str | None, str | None, bool | None]]:
    """Returns the set of contexts for each message type."""
    if descriptor is None:
        raise ValueError((descriptor, parent, field_name, unique))
    counts = {(descriptor.full_name, parent, field_name, unique)}
    for field in descriptor.fields:
        if field.type == FieldDescriptor.TYPE_MESSAGE:
            if set(field.message_type.fields_by_name.keys()) == {"key", "value"}:
                # Check for maps.
                logger.info(f"Found map: ({descriptor.full_name}, {field.name})")
                field_message_type = field.message_type.fields_by_name["value"].message_type
            else:
                field_message_type = field.message_type
            counts |= _get_message_contexts(
                field_message_type,
                parent=descriptor.full_name,
                field_name=field.name,
                unique=field.label != FieldDescriptor.LABEL_REPEATED,
            )
    return counts


def build_mappers() -> dict[Type[Message], Type]:
    """Creates ORM mapper classes for protocol buffer message types.

    Returns:
        Dict mapping protocol buffer message types to mapper classes.
    """
    mappers = {}
    parents = get_parents(dataset_pb2.Dataset)
    for message_type in sorted(parents, key=lambda x: x.DESCRIPTOR.name):
        logger.info(f"Building mapper for {message_type}")
        mappers[message_type] = build_mapper(message_type, parents=parents)
    return mappers


_FIELD_TYPES = {
    FieldDescriptor.TYPE_BOOL: Boolean,
    FieldDescriptor.TYPE_BYTES: LargeBinary,
    FieldDescriptor.TYPE_DOUBLE: Float,
    FieldDescriptor.TYPE_FLOAT: Float,
    FieldDescriptor.TYPE_INT32: Integer,
    FieldDescriptor.TYPE_INT64: Integer,
    FieldDescriptor.TYPE_STRING: Text,
}


def build_mapper(  # pylint: disable=too-many-branches
    message_type: Type[Message], parents: dict[Type[Message], list[tuple[Type[Message], str, bool]]]
) -> Type:
    """Creates a mapper class for a specific protocol buffer message type.

    Args:
        message_type: Protocol buffer message type.
        parents: Dict mapping message types to lists of (parent message type, field name, unique) tuples.

    Returns:
        Generated mapper class.
    """
    attrs = {
        "__tablename__": underscore(message_type.DESCRIPTOR.name),
        "id": Column(Integer, primary_key=True),
        "ord_schema_context": Column(Text, nullable=False),
    }
    attrs["__mapper_args__"] = {
        "polymorphic_on": attrs["ord_schema_context"],
        "polymorphic_identity": message_type.DESCRIPTOR.name,
        "with_polymorphic": "*",
    }
    add_key = False
    for parent_type, field_name, _ in parents[message_type]:
        if isinstance(getattr(parent_type(), field_name), Mapping):
            add_key = True
    if add_key:
        attrs["key"] = Column(Text)  # Map key.
    for field in message_type.DESCRIPTOR.fields:
        if field.name == "eic_masses":
            attrs[field.name] = Column(ARRAY(Float))
        elif field.name == "reaction_ids":
            attrs[field.name] = Column(ARRAY(Text))
        elif field.type == FieldDescriptor.TYPE_MESSAGE:
            kwargs = {}
            if field.label != FieldDescriptor.LABEL_REPEATED:
                kwargs["uselist"] = False
            # All relationships are to polymorphic child classes.
            child_class_name = f"_{message_type.DESCRIPTOR.name}{field.name.capitalize()}"
            attrs[field.name] = relationship(child_class_name, back_populates="parent", **kwargs)
        elif field.type == FieldDescriptor.TYPE_ENUM:
            attrs[field.name] = Column(Enum(*field.enum_type.values_by_name.keys(), name=field.enum_type.name))
        else:
            attrs[field.name] = Column(_FIELD_TYPES[field.type])
    if message_type == dataset_pb2.Dataset:
        # Make dataset IDs globally unique.
        attrs["dataset_id"] = Column(Text, nullable=False, unique=True)
    elif message_type == reaction_pb2.Reaction:
        # Make reaction IDs globally unique.
        attrs["reaction_id"] = Column(Text, nullable=False, unique=True)
        # Serialize and store the entire Reaction proto.
        attrs["proto"] = Column(LargeBinary, nullable=False)
        attrs["rdkit"] = relationship(f"_{message_type.DESCRIPTOR.name}RDKit", uselist=False)
    elif message_type in {reaction_pb2.Compound, reaction_pb2.ProductCompound}:
        attrs["rdkit"] = relationship(f"_{message_type.DESCRIPTOR.name}RDKit", uselist=False)
    elif message_type in {reaction_pb2.CompoundPreparation, reaction_pb2.CrudeComponent}:
        # Add foreign key to reaction.reaction_id.
        kwargs = {}
        if message_type == reaction_pb2.CrudeComponent:
            kwargs["nullable"] = False
        attrs["reaction_id"] = Column(Text, ForeignKey("reaction.reaction_id", ondelete="CASCADE"), **kwargs)
    logger.debug(f"Creating mapper {message_type.DESCRIPTOR.name}: {attrs}")
    mapper_class = type(message_type.DESCRIPTOR.name, (Base,), attrs)
    # Create polymorphic child classes.
    for parent_type, field_name, _ in parents[message_type]:
        foreign_table_name = underscore(parent_type.DESCRIPTOR.name)
        foreign_key = f"{foreign_table_name}.id"
        child_attrs = {
            "__mapper_args__": {"polymorphic_identity": f"{parent_type.DESCRIPTOR.name}.{field_name}"},
            # Use get() to avoid column conflicts; see
            # https://docs.sqlalchemy.org/en/14/orm/inheritance.html#resolving-column-conflicts.
            #
            # NOTE(skearnes): We are not enforcing unique constraints on this column; see the module docstring.
            f"{foreign_table_name}_id": mapper_class.__table__.c.get(
                f"{foreign_table_name}_id",
                Column(Integer, ForeignKey(foreign_key, ondelete="CASCADE")),
            ),
            "parent": relationship(parent_type.DESCRIPTOR.name, back_populates=field_name, uselist=False),
        }
        child_class_name = f"_{parent_type.DESCRIPTOR.name}{field_name.capitalize()}"
        logger.debug(f"Creating child mapper {child_class_name}: {child_attrs}")
        type(child_class_name, (mapper_class,), child_attrs)
    return mapper_class


_MESSAGE_TO_MAPPER: dict[Type[Message], Type] = build_mappers()
_MAPPER_TO_MESSAGE: dict[Type, Type[Message]] = {value: key for key, value in _MESSAGE_TO_MAPPER.items()}


class _MappersMeta(type):
    """Metaclass for Mappers; see https://stackoverflow.com/a/3155493."""

    def __getattr__(cls, item):
        return cls._MAPPERS[item]

    def __getitem__(cls, item):
        return cls._MAPPERS[item]


class Mappers(metaclass=_MappersMeta):
    """Container for generated mapper classes."""

    _MAPPERS: dict[str, Type] = {c.__name__: c for c in _MAPPER_TO_MESSAGE}


def from_proto(  # pylint: disable=too-many-branches
    message: Message, mapper: Optional[Type[Base]] = None, key: Optional[str] = None
) -> Base:
    """Converts a protobuf message into an ORM object.

    Args:
        message: Protobuf message.
        mapper: ORM mapper class. For top-level protos like Dataset and Reaction this can be left as None; it must
            be provided for Child subclasses to properly handle polymorphism.
        key: Map key (we store maps as rows of (key, value) tuples).

    Returns:
        ORM object.
    """
    if mapper is None:
        mapper = _MESSAGE_TO_MAPPER[type(message)]
    kwargs = {}
    if key is not None:
        kwargs["key"] = key
    if isinstance(message, reaction_pb2.Reaction):
        kwargs["proto"] = message.SerializeToString()
    for field, value in message.ListFields():
        if field.type == FieldDescriptor.TYPE_MESSAGE:
            field_mapper = getattr(mapper, field.name).mapper.class_
            if isinstance(value, Mapping):
                kwargs[field.name] = [from_proto(v, mapper=field_mapper, key=k) for k, v in value.items()]
            elif field.label == FieldDescriptor.LABEL_REPEATED:
                kwargs[field.name] = [from_proto(v, mapper=field_mapper) for v in value]
            else:
                kwargs[field.name] = from_proto(value, mapper=field_mapper)
        elif field.type == FieldDescriptor.TYPE_ENUM:
            kwargs[field.name] = field.enum_type.values_by_number[value].name
        else:
            kwargs[field.name] = value
    if isinstance(message, reaction_pb2.Reaction):
        field_mapper = getattr(mapper, "rdkit").mapper.class_
        try:
            reaction_smiles = message_helpers.get_reaction_smiles(message, generate_if_missing=True)
            assert reaction_smiles is not None  # Type hint.
            reaction_smiles = reaction_smiles.split()[0]  # Handle CXSMILES.
            kwargs["rdkit"] = field_mapper(reaction_smiles=reaction_smiles)
        except ValueError:
            pass
    elif isinstance(message, (reaction_pb2.Compound, reaction_pb2.ProductCompound)):
        # Add RDKit cartridge functionality.
        field_mapper = getattr(mapper, "rdkit").mapper.class_
        try:
            kwargs["rdkit"] = field_mapper(smiles=message_helpers.smiles_from_compound(message))
        except ValueError:
            pass
    return mapper(**kwargs)


def to_proto(base: Base) -> Message:
    """Converts an ORM object into a protobuf message.

    Args:
        base: ORM object.

    Returns:
        Protobuf message.
    """
    kwargs = {}
    try:
        proto = _MAPPER_TO_MESSAGE[type(base)]
    except KeyError:
        proto = _MAPPER_TO_MESSAGE[type(base).__bases__[0]]
    assert issubclass(proto, Message)
    for field in proto.DESCRIPTOR.fields:
        value = getattr(base, field.name)
        if isinstance(value, list) and len(value) == 0:
            continue
        if isinstance(value, list) and isinstance(value[0], Base):
            if hasattr(value[0], "key"):
                kwargs[field.name] = {v.key: to_proto(v) for v in value}
            else:
                kwargs[field.name] = [to_proto(v) for v in value]
        elif isinstance(value, Base):
            kwargs[field.name] = to_proto(value)
        else:
            kwargs[field.name] = value
    return proto(**kwargs)



# Copyright 2022 Open Reaction Database Project Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for ord_schema.orm.mappers."""
import os
import pytest

from ord_schema.message_helpers import load_message
from ord_schema.orm.mappers import from_proto, to_proto
from ord_schema.proto.dataset_pb2 import Dataset


@pytest.mark.parametrize(
    "filename",
    (
        os.path.join(os.path.dirname(__file__), "testdata", "empty.pbtxt"),
        os.path.join(os.path.dirname(__file__), "testdata", "full.pbtxt"),
        os.path.join(os.path.dirname(__file__), "testdata", "ord-nielsen-example.pbtxt"),
    ),
)
def test_round_trip(filename):
    dataset = load_message(filename, Dataset)
    assert dataset == to_proto(from_proto(dataset))



# Copyright 2020 Open Reaction Database Project Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for ord_schema.validations."""

import sys
import warnings

import pytest
from google.protobuf import text_format

from ord_schema import validations
from ord_schema.proto import dataset_pb2
from ord_schema.proto import reaction_pb2


# pylint: disable=too-many-public-methods


@pytest.fixture(autouse=True)
def setup():
    # Redirect warning messages to stdout so they can be filtered from the other test output.
    original_showwarning = warnings.showwarning

    # pylint: disable=too-many-arguments
    def _showwarning(message, category, filename, lineno, file=None, line=None):
        del file  # Unused.
        original_showwarning(
            message=message, category=category, filename=filename, lineno=lineno, file=sys.stdout, line=line
        )

    # pylint: enable=too-many-arguments
    warnings.showwarning = _showwarning
    yield
    # Restore the original showwarning.
    warnings.showwarning = original_showwarning


def _run_validation(message, **kwargs):
    original = type(message)()
    original.CopyFrom(message)
    output = validations.validate_message(message, **kwargs)
    # Verify that `message` is unchanged by the validation process.
    assert original == message
    return output


@pytest.mark.parametrize(
    "message",
    (
        reaction_pb2.ReactionNotes(),
        reaction_pb2.StirringConditions(type="UNSPECIFIED"),
        reaction_pb2.ReactionNotes(safety_notes=""),
    ),
)
def test_is_empty(message):
    assert validations.is_empty(message)


@pytest.mark.parametrize(
    "message",
    (
        reaction_pb2.StirringConditions(type="STIR_BAR"),
        reaction_pb2.ReactionNotes(is_heterogeneous=False),
        reaction_pb2.ReactionNotes(is_heterogeneous=True),
    ),
)
def test_is_not_empty(message):
    assert not validations.is_empty(message)


@pytest.mark.parametrize(
    "message",
    (
        reaction_pb2.Volume(value=15.0, units=reaction_pb2.Volume.MILLILITER),
        reaction_pb2.Time(value=24, units=reaction_pb2.Time.HOUR),
        reaction_pb2.Mass(value=32.1, units=reaction_pb2.Mass.GRAM),
    ),
)
def test_units(message):
    output = _run_validation(message)
    assert len(output.errors) == 0
    assert len(output.warnings) == 0


@pytest.mark.parametrize(
    "message,expected",
    (
        (reaction_pb2.Volume(value=-15.0, units=reaction_pb2.Volume.MILLILITER), "non-negative"),
        (reaction_pb2.Time(value=-24, units=reaction_pb2.Time.HOUR), "non-negative"),
        (reaction_pb2.Mass(value=-32.1, units=reaction_pb2.Mass.GRAM), "non-negative"),
        (reaction_pb2.FlowRate(value=5), "units"),
        (reaction_pb2.Temperature(value=-5, units="KELVIN"), "between"),
        (reaction_pb2.Temperature(value=-500, units="CELSIUS"), "between"),
    ),
)
def test_units_should_fail(message, expected):
    with pytest.raises(validations.ValidationError, match=expected):
        _run_validation(message)


def test_orcid():
    message = reaction_pb2.Person(orcid="0000-0001-2345-678X")
    output = _run_validation(message)
    assert len(output.errors) == 0
    assert len(output.warnings) == 0


def test_orcid_should_fail():
    message = reaction_pb2.Person(orcid="abcd-0001-2345-678X")
    with pytest.raises(validations.ValidationError, match="Invalid"):
        _run_validation(message)


@pytest.mark.parametrize("email", ["ord+test@gmail.com", "student@alumni.mit.edu", "hypen-ated@gmail.com"])
def test_email(email):
    message = reaction_pb2.Person(email=email)
    output = _run_validation(message)
    assert len(output.errors) == 0
    assert len(output.warnings) == 0


@pytest.mark.parametrize("email", ["bad&character@gmail.com", "not-an-email", "bad@domain"])
def test_email_should_fail(email):
    message = reaction_pb2.Person(email=email)
    with pytest.raises(validations.ValidationError, match="Invalid"):
        _run_validation(message)


def test_synthesized_compound():
    message = reaction_pb2.CompoundPreparation(type="SYNTHESIZED", reaction_id="dummy_reaction_id")
    output = _run_validation(message)
    assert len(output.errors) == 0
    assert len(output.warnings) == 0
    message = reaction_pb2.CompoundPreparation(type="NONE", reaction_id="dummy_reaction_id")
    with pytest.raises(validations.ValidationError, match="only .* when SYNTHESIZED"):
        _run_validation(message)


def test_unmeasured_amount():
    message = reaction_pb2.ReactionInput()
    message.components.add().CopyFrom(
        reaction_pb2.Compound(
            identifiers=[dict(type="SMILES", value="c1ccccc1")],
            amount=dict(unmeasured=dict(type="SATURATED")),
        )
    )
    output = _run_validation(message)
    assert len(output.warnings) == 1
    message.components.add().CopyFrom(reaction_pb2.Compound(identifiers=[dict(type="SMILES", value="O")]))
    output = _run_validation(message)
    assert len(output.errors) == 0
    assert len(output.warnings) == 0


def test_crude_component():
    message = reaction_pb2.CrudeComponent()
    with pytest.raises(validations.ValidationError, match="reaction_id"):
        _run_validation(message)
    message = reaction_pb2.CrudeComponent(reaction_id="my_reaction_id")
    with pytest.raises(validations.ValidationError, match="amount"):
        _run_validation(message)
    message = reaction_pb2.CrudeComponent(reaction_id="my_reaction_id", has_derived_amount=False)
    with pytest.raises(validations.ValidationError, match="amount"):
        _run_validation(message)
    message = reaction_pb2.CrudeComponent(reaction_id="my_reaction_id", has_derived_amount=True)
    output = _run_validation(message)
    assert len(output.errors) == 0
    assert len(output.warnings) == 0


def test_product_measurement():
    message = reaction_pb2.ProductMeasurement(analysis_key="my_analysis", type="YIELD", float_value=dict(value=60))
    output = _run_validation(message)
    assert "percentage" in output.warnings[0]
    message = reaction_pb2.ProductMeasurement(analysis_key="my_analysis", type="AREA", string_value="35.221")
    with pytest.raises(validations.ValidationError, match="numeric"):
        _run_validation(message)
    message = reaction_pb2.ProductMeasurement(
        analysis_key="my_analysis", type="YIELD", percentage=dict(value=60), selectivity=dict(type="EE")
    )
    with pytest.raises(validations.ValidationError, match="selectivity"):
        _run_validation(message)


def test_reaction():
    message = reaction_pb2.Reaction()
    with pytest.raises(validations.ValidationError, match="reaction input"):
        _run_validation(message)


def test_reaction_smiles():
    message = reaction_pb2.Reaction()
    message.identifiers.add(value="C>>C", type="REACTION_SMILES")
    output = _run_validation(message)
    assert len(output.errors) == 0
    assert len(output.warnings) == 0
    # Now disable the exception for reaction SMILES only.
    options = validations.ValidationOptions()
    options.allow_reaction_smiles_only = False
    with pytest.raises(validations.ValidationError, match="reaction input"):
        _run_validation(message, options=options)


def test_bad_reaction_smiles():
    message = reaction_pb2.Reaction()
    message.identifiers.add(value="test", type="REACTION_SMILES")
    with pytest.raises(validations.ValidationError, match="requires at least two > characters"):
        _run_validation(message)


@pytest.mark.parametrize(
    "workup_text",
    (
        "type: ALIQUOT amount {mass {value: 1.0 units: GRAM}}",
        "type: ADDITION input {components {"
        'identifiers {value: "CCO" type: SMILES} '
        "amount {mass {value: 10.0 units: GRAM}}}}",
    ),
)
def test_reaction_workup(workup_text):
    message = text_format.Parse(workup_text, reaction_pb2.ReactionWorkup())
    output = _run_validation(message)
    assert len(output.errors) == 0
    assert len(output.warnings) == 0


@pytest.mark.parametrize(
    "workup_text,error_msg",
    (
        ("type: ALIQUOT", "missing volume/mass"),
        (
            "type: ADDITION amount {mass {value: 1.0 units: GRAM}} "
            'input {components {identifiers {value: "CCO" type: SMILES} '
            "amount {mass {value: 10.0 units: GRAM}}}}",
            "should only be specified",
        ),
    ),
)
def test_bad_reaction_workup(workup_text, error_msg):
    message = text_format.Parse(workup_text, reaction_pb2.ReactionWorkup())
    output = _run_validation(message)
    assert len(output.warnings) == 1
    assert error_msg in output.warnings[0]


# pylint: disable=too-many-statements
def test_reaction_recursive():
    message = reaction_pb2.Reaction()
    # Reactions must have at least one input
    with pytest.raises(validations.ValidationError, match="reaction input"):
        _run_validation(message, recurse=False)
    dummy_input = message.inputs["dummy_input"]
    # Reactions must have at least one outcome
    with pytest.raises(validations.ValidationError, match="reaction outcome"):
        _run_validation(message, recurse=False)
    outcome = message.outcomes.add()
    output = _run_validation(message, recurse=False)
    assert len(output.errors) == 0
    assert len(output.warnings) == 0
    # Inputs must have at least one component
    with pytest.raises(validations.ValidationError, match="component"):
        _run_validation(message)
    dummy_component = dummy_input.components.add()
    # Components must have at least one identifier
    with pytest.raises(validations.ValidationError, match="identifier"):
        _run_validation(message)
    dummy_component.identifiers.add(type="CUSTOM")
    # Custom identifiers must have details specified
    with pytest.raises(validations.ValidationError, match="details"):
        _run_validation(message)
    dummy_component.identifiers[0].details = "custom_identifier"
    dummy_component.identifiers[0].value = "custom_value"
    # Components of reaction inputs must have a defined amount
    with pytest.raises(validations.ValidationError, match="require an amount"):
        _run_validation(message)
    dummy_component.amount.mass.value = 1
    dummy_component.amount.mass.units = reaction_pb2.Mass.GRAM
    # Reactions don't require defined products or conversion because they
    # can be used to prepare crude for a second Reaction
    output = _run_validation(message)
    assert len(output.errors) == 0
    assert len(output.warnings) == 1
    outcome.conversion.value = 75
    # If converseions are defined, must have limiting reagent flag
    with pytest.raises(validations.ValidationError, match="is_limiting"):
        _run_validation(message)
    dummy_component.is_limiting = True
    output = _run_validation(message)
    assert len(output.errors) == 0
    assert len(output.warnings) == 0

    # If a measurement uses an internal standard, a component must have
    # an INTERNAL_STANDARD reaction role
    outcome.analyses["dummy_analysis"].CopyFrom(reaction_pb2.Analysis(type="CUSTOM", details="test"))
    product = outcome.products.add(identifiers=[dict(type="SMILES", value="c1ccccc1")])
    product.measurements.add(
        analysis_key="dummy_analysis", type="YIELD", percentage=dict(value=75), uses_internal_standard=True
    )
    with pytest.raises(validations.ValidationError, match="INTERNAL_STANDARD"):
        _run_validation(message)
    # Assigning internal standard role to input should resolve the error
    message_input_istd = reaction_pb2.Reaction()
    message_input_istd.CopyFrom(message)
    message_input_istd.inputs["dummy_input"].components[0].reaction_role = reaction_pb2.ReactionRole.INTERNAL_STANDARD
    output = _run_validation(message_input_istd)
    assert len(output.errors) == 0
    assert len(output.warnings) == 0
    # Assigning internal standard role to workup should resolve the error
    message_workup_istd = reaction_pb2.Reaction()
    message_workup_istd.CopyFrom(message)
    workup = message_workup_istd.workups.add(type="CUSTOM", details="test")
    istd = workup.input.components.add()
    istd.identifiers.add(type="SMILES", value="CCO")
    istd.amount.mass.value = 1
    istd.amount.mass.units = reaction_pb2.Mass.GRAM
    istd.reaction_role = reaction_pb2.ReactionRole.INTERNAL_STANDARD
    output = _run_validation(message_workup_istd)
    assert len(output.errors) == 0
    assert len(output.warnings) == 0


def test_reaction_recursive_noraise_on_error():
    message = reaction_pb2.Reaction()
    message.inputs["dummy_input"].components.add()
    output = _run_validation(message, raise_on_error=False)
    expected = [
        'Reaction.inputs["dummy_input"].components[0]: ' "Compounds must have at least one identifier",
        "Reaction: Reactions should have at least 1 reaction outcome",
        "Reaction: All reaction input components require an amount",
    ]
    assert output.errors == expected
    assert len(output.warnings) == 1


def test_datetimes():
    message = reaction_pb2.ReactionProvenance()
    message.experiment_start.value = "2020-01-02"
    message.record_created.time.value = "2020-01-01"
    message.record_created.person.name = "test"
    message.record_created.person.email = "test@example.com"
    with pytest.raises(validations.ValidationError, match="after"):
        _run_validation(message)
    message.record_created.time.value = "2020-01-03"
    output = _run_validation(message)
    assert len(output.errors) == 0
    assert len(output.warnings) == 0


def test_reaction_id():
    message = reaction_pb2.Reaction()
    _ = message.inputs["test"]
    message.outcomes.add()
    message.reaction_id = "ord-c0bbd41f095a44a78b6221135961d809"
    options = validations.ValidationOptions(validate_ids=True)
    output = _run_validation(message, recurse=False, options=options)
    assert len(output.errors) == 0
    assert len(output.warnings) == 0


@pytest.mark.parametrize(
    "reaction_id",
    (
        "ord-c0bbd41f095a4",
        "ord-c0bbd41f095a4c0bbd41f095a4c0bbd41f095a4",
        "foo-c0bbd41f095a44a78b6221135961d809",
        "ord-C0BBD41F095A44A78B6221135961D809",
        "ord-h0bbd41f095a44a78b6221135961d809",
        "ord-notARealId",
        "",
    ),
)
def test_bad_reaction_id(reaction_id):
    message = reaction_pb2.Reaction(reaction_id=reaction_id)
    _ = message.inputs["test"]
    message.outcomes.add()
    options = validations.ValidationOptions(validate_ids=True)
    with pytest.raises(validations.ValidationError, match="malformed"):
        _run_validation(message, recurse=False, options=options)


def test_missing_provenance():
    message = reaction_pb2.Reaction()
    _ = message.inputs["test"]
    message.outcomes.add()
    options = validations.ValidationOptions(require_provenance=True)
    with pytest.raises(validations.ValidationError, match="requires provenance"):
        _run_validation(message, recurse=False, options=options)


def test_data():
    message = reaction_pb2.Data()
    with pytest.raises(validations.ValidationError, match="requires one of"):
        _run_validation(message)
    message.bytes_value = b"test data"
    with pytest.raises(validations.ValidationError, match="format is required"):
        _run_validation(message)
    message.string_value = "test data"
    output = _run_validation(message)
    assert len(output.errors) == 0
    assert len(output.warnings) == 0


def test_dataset_bad_reaction_id():
    message = dataset_pb2.Dataset(reaction_ids=["foo"])
    options = validations.ValidationOptions(validate_ids=True)
    with pytest.raises(validations.ValidationError, match="malformed"):
        _run_validation(message, options=options)


def test_dataset_records_and_ids():
    message = dataset_pb2.Dataset(
        reactions=[reaction_pb2.Reaction()],
        reaction_ids=["ord-c0bbd41f095a44a78b6221135961d809"],
    )
    options = validations.ValidationOptions(validate_ids=True)
    with pytest.raises(validations.ValidationError, match="not both"):
        _run_validation(message, recurse=False, options=options)


def test_dataset_bad_id():
    message = dataset_pb2.Dataset(reactions=[reaction_pb2.Reaction()], dataset_id="foo")
    options = validations.ValidationOptions(validate_ids=True)
    with pytest.raises(validations.ValidationError, match="malformed"):
        _run_validation(message, recurse=False, options=options)


def test_dataset_example():
    message = dataset_pb2.DatasetExample()
    with pytest.raises(validations.ValidationError, match="description is required"):
        _run_validation(message)
    message.description = "test example"
    with pytest.raises(validations.ValidationError, match="url is required"):
        _run_validation(message)
    message.url = "example.com"
    with pytest.raises(validations.ValidationError, match="created is required"):
        _run_validation(message)
    message.created.time.value = "11 am"
    message.created.person.name = "test"
    message.created.person.email = "test@example.com"
    output = _run_validation(message)
    assert len(output.errors) == 0
    assert len(output.warnings) == 0


def test_dataset_crossreferences():
    message = dataset_pb2.Dataset()
    reaction1 = message.reactions.add()
    reaction2 = message.reactions.add()
    reaction3 = message.reactions.add()
    # Minimal reaction 1
    dummy_input = reaction1.inputs["dummy_input"]
    reaction1.outcomes.add()
    dummy_component = dummy_input.components.add()
    dummy_component.identifiers.add(type="CUSTOM")
    dummy_component.identifiers[0].details = "custom_identifier"
    dummy_component.identifiers[0].value = "custom_value"
    dummy_component.amount.mass.value = 1
    dummy_component.amount.mass.units = reaction_pb2.Mass.GRAM
    reaction2.CopyFrom(reaction1)
    reaction3.CopyFrom(reaction1)
    # It is not required to cross-reference SYNTHESIZED compounds.
    dummy_component.preparations.add(type="SYNTHESIZED")
    output = _run_validation(message)
    assert len(output.errors) == 0
    assert len(output.warnings) == 3
    # References must refer to ID within this dataset.
    dummy_component.preparations[0].reaction_id = "placeholder_id"
    with pytest.raises(validations.ValidationError, match="undefined reaction_ids"):
        _run_validation(message)
    # Self-references aren't allowed
    reaction1.reaction_id = "placeholder_id"
    with pytest.raises(validations.ValidationError, match="own ID"):
        _run_validation(message)
    # Valid reference
    reaction1.reaction_id = ""
    reaction2.reaction_id = "placeholder_id"
    output = _run_validation(message)
    assert len(output.errors) == 0
    assert len(output.warnings) == 3
    # Duplicate IDs not allowed
    reaction3.reaction_id = "placeholder_id"
    with pytest.raises(validations.ValidationError, match="same IDs"):
        _run_validation(message)
    reaction3.reaction_id = ""
    # Crude component also needs valid IDs
    dummy_input.crude_components.add(reaction_id="crude-making step", has_derived_amount=True)
    with pytest.raises(validations.ValidationError, match="undefined reaction_ids"):
        _run_validation(message)
    reaction3.reaction_id = "crude-making step"
    output = _run_validation(message)
    assert len(output.errors) == 0
    assert len(output.warnings) == 3



# Copyright 2022 Open Reaction Database Project Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Functions for creating/managing the PostgreSQL database."""
import time
import os
from unittest.mock import patch

from sqlalchemy import cast, delete, func, text, update
from sqlalchemy.engine import Engine
from sqlalchemy.exc import OperationalError
from sqlalchemy.orm import Session

from ord_schema.logging import get_logger
from ord_schema.orm.mappers import Base, Mappers, from_proto
from ord_schema.orm.rdkit_mappers import CString, FingerprintType, RDKitMol, RDKitReaction
from ord_schema.proto import dataset_pb2

logger = get_logger(__name__)


def get_connection_string(
    database: str, username: str, password: str, host: str = "localhost", port: int = 5432
) -> str:
    """Creates an SQLAlchemy connection string."""
    return f"postgresql://{username}:{password}@{host}:{port}/{database}"


def prepare_database(engine: Engine) -> bool:
    """Prepares the database and creates the ORM table structure.

    Args:
        engine: SQLAlchemy Engine.

    Returns:
        Whether the RDKit PostgreSQL cartridge is installed.
    """
    with engine.begin() as connection:
        try:
            connection.execute(text("CREATE EXTENSION IF NOT EXISTS tsm_system_rows"))  # For random sampling.
        except OperationalError:
            logger.warning("tsm_system_rows cartridge is not installed; random sampling will be disabled")
    with engine.begin() as connection:
        connection.execute(text("CREATE SCHEMA IF NOT EXISTS rdkit"))
        try:
            connection.execute(text("CREATE EXTENSION IF NOT EXISTS rdkit WITH SCHEMA rdkit"))
            rdkit_cartridge = True
        except OperationalError:
            logger.warning("RDKit PostgreSQL cartridge is not installed; structure search will be disabled")
            connection.execute(text("CREATE EXTENSION IF NOT EXISTS btree_gist WITH SCHEMA rdkit"))
            rdkit_cartridge = False
    with patch.dict(os.environ, {"ORD_POSTGRES_RDKIT": "1" if rdkit_cartridge else "0"}):
        Base.metadata.create_all(engine)
    return rdkit_cartridge


def add_dataset(dataset: dataset_pb2.Dataset, session: Session) -> None:
    """Adds a dataset to the database."""
    logger.info(f"Adding dataset {dataset.dataset_id}")
    start = time.time()
    mapped_dataset = from_proto(dataset)
    logger.info(f"from_proto() took {time.time() - start}s")
    start = time.time()
    session.add(mapped_dataset)
    logger.info(f"session.add() took {time.time() - start}s")


def delete_dataset(dataset_id: str, session: Session) -> None:
    """Deletes a dataset from the database."""
    logger.info(f"Deleting dataset {dataset_id}")
    start = time.time()
    session.execute(delete(Mappers.Dataset).where(Mappers.Dataset.dataset_id == dataset_id))
    logger.info(f"delete took {time.time() - start}s")


def add_rdkit(session: Session) -> None:
    """Adds RDKit PostgreSQL cartridge data to any null-valued cells in the structure table."""
    logger.info("Populating RDKit reaction columns")
    assert hasattr(RDKitReaction, "__table__")  # Type hint.
    table = RDKitReaction.__table__
    start = time.time()
    session.execute(
        update(table)
        .where(table.c.reaction.is_(None))
        .values(reaction=func.rdkit.reaction_from_smiles(cast(table.c.reaction_smiles, CString)))
    )
    logger.info(f"Adding reaction took {time.time() - start}s")
    logger.info("Populating RDKit mol columns")
    assert hasattr(RDKitMol, "__table__")  # Type hint.
    table = RDKitMol.__table__
    start = time.time()
    session.execute(
        update(table).where(table.c.mol.is_(None)).values(mol=func.rdkit.mol_from_smiles(cast(table.c.smiles, CString)))
    )
    logger.info(f"Adding mol took {time.time() - start}s")
    for fp_type in FingerprintType:
        start = time.time()
        column = fp_type.name.lower()
        session.execute(
            update(table)
            .where(getattr(table.c, column).is_(None), table.c.mol.is_not(None))
            .values(**{column: fp_type(table.c.mol)})
        )
        logger.info(f"Adding {fp_type} took {time.time() - start}s")



# Copyright 2022 Open Reaction Database Project Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for ord_schema.orm.add_datasets."""
import os

import docopt
import pytest
from sqlalchemy import create_engine
from sqlalchemy.exc import IntegrityError
from testing.postgresql import Postgresql

from ord_schema.orm import add_datasets
from ord_schema.orm.database import prepare_database


def test_main():
    with Postgresql() as postgres:
        engine = create_engine(postgres.url(), future=True)
        assert prepare_database(engine)  # Requires RDKit.
        argv = [
            "--url",
            postgres.url(),
            "--pattern",
            os.path.join(os.path.dirname(__file__), "testdata", "ord-nielsen-example.pbtxt"),
        ]
        add_datasets.main(**docopt.docopt(add_datasets.__doc__, argv))
        with pytest.raises(IntegrityError, match="violates unique constraint"):
            add_datasets.main(**docopt.docopt(add_datasets.__doc__, argv))
        argv.append("--overwrite")
        add_datasets.main(**docopt.docopt(add_datasets.__doc__, argv))
